from __future__ import annotations

import collections
from typing import TYPE_CHECKING, Literal

import discord

from pylav.filters import ChannelMix, Distortion, Equalizer, Karaoke, LowPass, Rotation, Timescale, Vibrato, Volume
from pylav.filters.tremolo import Tremolo
from pylav.utils import Segment

if TYPE_CHECKING:
    from pylav.node import Node
    from pylav.player import Player
    from pylav.tracks import Track


class Event:
    """The base for all Lavalink events."""


class QueueEndEvent(Event):
    """This event is dispatched when there are no more songs in the queue.

    Attributes
    ----------
    player: :class:`Player`
        The player that has no more songs in queue.

    Parameters
    ----------
    player: :class:`Player`
        The player that has no more songs in queue.
    """

    __slots__ = ("player",)

    def __init__(self, player: Player):
        self.player = player


class TrackStuckEvent(Event):
    """This event is dispatched when the currently playing track is stuck.
    This normally has something to do with the stream you are playing
    and not Lavalink itself.

    Attributes
    ----------
    player: :class:`Player`
        The player that has the playing track being stuck.
    track: :class:`Track`
        The track is stuck from playing.
    threshold: :class:`int`
        The amount of time the track had while being stuck.
    node: :class:`Node`
        The node that the track is stuck on.

    Parameters
    ----------
    player: :class:`Player`
        The player that has the playing track being stuck.
    track: :class:`Track`
        The track is stuck from playing.
    threshold: :class:`int`
        The amount of time the track had while being stuck.
    node: :class:`Node`
        The node that the stuck track is playing on.
    """

    __slots__ = ("player", "track", "threshold", "node")

    def __init__(self, player: Player, track: Track, threshold: float, node: Node):
        self.player = player
        self.track = track
        self.threshold = threshold
        self.node = node


class TrackExceptionEvent(Event):
    """This event is dispatched when an exception occurs while playing a track.

    Attributes
    ----------
    player: :class:`Player`
        The player that had the exception occur while playing a track.
    track: :class:`Track`
        The track that had the exception while playing.
    exception: :class:`Exception`
        The type of exception that the track had while playing.
    node: :class:`Node`
        The node that the exception occurred on.

    Parameters
    ----------
    player: :class:`Player`
        The player that had the exception occur while playing a track.
    track: :class:`Track`
        The track that had the exception while playing.
    exception: :class:`Exception`
        The type of exception that the track had while playing.
    node: :class:`Node`
        The node that the exception occurred on.
    """

    __slots__ = ("player", "track", "exception", "node")

    def __init__(self, player: Player, track: Track, exception: Exception, node: Node):
        self.player = player
        self.track = track
        self.exception = exception
        self.node = node


class TrackEndEvent(Event):
    """This event is dispatched when the player finished playing a track.

    Attributes
    ----------
    player: :class:`Player`
        The player that finished playing a track.
    track: :class:`Track`
        The track that finished playing.
    reason: :class:`str`
        The reason why the track stopped playing.
    node: :class:`Node`
        The node that the track finished playing on.

    Parameters
    ----------
    player: :class:`Player`
        The player that finished playing a track.
    track: :class:`Track`
        The track that finished playing.
    reason: :class:`str`
        The reason why the track stopped playing.
    node: :class:`Node`
        The node that the track finished playing on.
    """

    __slots__ = ("player", "track", "reason", "node")

    def __init__(self, player: Player, track: Track, reason: str, node: Node):
        self.player = player
        self.track = track
        self.reason = reason
        self.node = node


class TrackStartEvent(Event):
    """This event is dispatched when the player starts to play a track.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    url: :class:`str`
        The url of the track that started playing.
    identifier: :class:`str`
        The identifier of the track that started playing.
    duration: :class:`int`
        The duration of the track that started playing.
    title: :class:`str`
        The title of the track that started playing.
    author: :class:`str`
        The author of the track that started playing.
    node: :class:`Node`
        The node that the track started playing on.

    Parameters
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    node: :class:`Node`
        The node that the track started playing on.
    """

    __slots__ = ("player", "track", "url", "identifier", "duration", "title", "author", "node")

    def __init__(self, player: Player, track: Track, node: Node):
        self.player = player
        self.track = track
        self.url = track.uri
        self.identifier = track.identifier
        self.duration = track.duration
        self.title = track.title
        self.author = track.author
        self.node = node


class TrackAutoPlayEvent(Event):
    """This event is dispatched when the player starts to play a track.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.

    Parameters
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing
    """

    __slots__ = ("player", "track")

    def __init__(self, player: Player, track: Track):
        self.player = player
        self.track = track


class TrackResumedEvent(Event):
    """This event is dispatched when the player resumes playing a track.

    Attributes
    ----------
    player: :class:`Player`
        The player that resumed playing a track.
    track: :class:`Track`
        The track that resumed playing.
    requester: :class:`discord.Member`
        The member that requested the track to resume playing.

    Parameters
    ----------
    player: :class:`Player`
        The player that resumed playing a track.
    track: :class:`Track`
        The track that resumed playing.
    requester: :class:`discord.Member`
        The member that requested the track to resume playing.
    """

    __slots__ = ("player", "track", "requester")

    def __init__(self, player: Player, track: Track, requester: discord.Member):
        self.player = player
        self.track = track
        self.requester = requester


class TrackStartYouTubeEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a YouTube track.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    url: :class:`str`
        The url of the track that started playing.
    identifier: :class:`str`
        The identifier of the track that started playing.
    duration: :class:`int`
        The duration of the track that started playing.
    title: :class:`str`
        The title of the track that started playing.
    author: :class:`str`
        The author of the track that started playing.
    node: :class:`Node`
        The node that the track started playing on.

    Parameters
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    node: :class:`Node`
        The node that the track started playing on.
    """

    __slots__ = ("player", "track", "url", "identifier", "duration", "title", "author", "node")

    def __init__(self, player: Player, track: Track, node: Node):
        super().__init__(player, track, node)
        self.url = track.uri
        self.identifier = track.identifier
        self.duration = track.duration
        self.title = track.title
        self.author = track.author


class TrackStartClypitEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Clyp.it track.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    url: :class:`str`
        The url of the track that started playing.
    identifier: :class:`str`
        The identifier of the track that started playing.
    duration: :class:`int`
        The duration of the track that started playing.
    title: :class:`str`
        The title of the track that started playing.
    author: :class:`str`
        The author of the track that started playing.
    node: :class:`Node`
        The node that the track started playing on.

    Parameters
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    node: :class:`Node`
        The node that the track started playing on.
    """

    __slots__ = ("player", "track", "url", "identifier", "duration", "title", "author", "node")

    def __init__(self, player: Player, track: Track, node: Node):
        super().__init__(player, track, node)
        self.url = track.uri
        self.identifier = track.identifier
        self.duration = track.duration
        self.title = track.title
        self.author = track.author


class TrackStartGetYarnEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a GetYarn track.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    url: :class:`str`
        The url of the track that started playing.
    identifier: :class:`str`
        The identifier of the track that started playing.
    duration: :class:`int`
        The duration of the track that started playing.
    title: :class:`str`
        The title of the track that started playing.
    author: :class:`str`
        The author of the track that started playing.
    node: :class:`Node`
        The node that the track started playing on.

    Parameters
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    node: :class:`Node`
        The node that the track started playing on.
    """

    __slots__ = ("player", "track", "url", "identifier", "duration", "title", "author", "node")

    def __init__(self, player: Player, track: Track, node: Node):
        super().__init__(player, track, node)
        self.url = track.uri
        self.identifier = track.identifier
        self.duration = track.duration
        self.title = track.title
        self.author = track.author


class TrackStartMixCloudEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a MixCloud track.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    url: :class:`str`
        The url of the track that started playing.
    identifier: :class:`str`
        The identifier of the track that started playing.
    duration: :class:`int`
        The duration of the track that started playing.
    title: :class:`str`
        The title of the track that started playing.
    author: :class:`str`
        The author of the track that started playing.
    node: :class:`Node`
        The node that the track started playing on.

    Parameters
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    node: :class:`Node`
        The node that the track started playing on.
    """

    __slots__ = ("player", "track", "url", "identifier", "duration", "title", "author", "node")

    def __init__(self, player: Player, track: Track, node: Node):
        super().__init__(player, track, node)
        self.url = track.uri
        self.identifier = track.identifier
        self.duration = track.duration
        self.title = track.title
        self.author = track.author


class TrackStartOCRMixEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a OCR Mix track.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    url: :class:`str`
        The url of the track that started playing.
    identifier: :class:`str`
        The identifier of the track that started playing.
    duration: :class:`int`
        The duration of the track that started playing.
    title: :class:`str`
        The title of the track that started playing.
    author: :class:`str`
        The author of the track that started playing.
    node: :class:`Node`
        The node that the track started playing on.

    Parameters
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    node: :class:`Node`
        The node that the track started playing on.
    """

    __slots__ = ("player", "track", "url", "identifier", "duration", "title", "author", "node")

    def __init__(self, player: Player, track: Track, node: Node):
        super().__init__(player, track, node)
        self.url = track.uri
        self.identifier = track.identifier
        self.duration = track.duration
        self.title = track.title
        self.author = track.author


class TrackStartPornHubEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Pornhub track.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    url: :class:`str`
        The url of the track that started playing.
    identifier: :class:`str`
        The identifier of the track that started playing.
    duration: :class:`int`
        The duration of the track that started playing.
    title: :class:`str`
        The title of the track that started playing.
    author: :class:`str`
        The author of the track that started playing.
    node: :class:`Node`
        The node that the track started playing on.

    Parameters
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    node: :class:`Node`
        The node that the track started playing on.
    """

    __slots__ = ("player", "track", "url", "identifier", "duration", "title", "author", "node")

    def __init__(self, player: Player, track: Track, node: Node):
        super().__init__(player, track, node)
        self.url = track.uri
        self.identifier = track.identifier
        self.duration = track.duration
        self.title = track.title
        self.author = track.author


class TrackStartRedditEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Reddit track.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    url: :class:`str`
        The url of the track that started playing.
    identifier: :class:`str`
        The identifier of the track that started playing.
    duration: :class:`int`
        The duration of the track that started playing.
    title: :class:`str`
        The title of the track that started playing.
    author: :class:`str`
        The author of the track that started playing.
    node: :class:`Node`
        The node that the track started playing on.

    Parameters
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    node: :class:`Node`
        The node that the track started playing on.
    """

    __slots__ = ("player", "track", "url", "identifier", "duration", "title", "author", "node")

    def __init__(self, player: Player, track: Track, node: Node):
        super().__init__(player, track, node)
        self.url = track.uri
        self.identifier = track.identifier
        self.duration = track.duration
        self.title = track.title
        self.author = track.author


class TrackStartSoundgasmEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Soundgasm track.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    url: :class:`str`
        The url of the track that started playing.
    identifier: :class:`str`
        The identifier of the track that started playing.
    duration: :class:`int`
        The duration of the track that started playing.
    title: :class:`str`
        The title of the track that started playing.
    author: :class:`str`
        The author of the track that started playing.
    node: :class:`Node`
        The node that the track started playing on.

    Parameters
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    node: :class:`Node`
        The node that the track started playing on.
    """

    __slots__ = ("player", "track", "url", "identifier", "duration", "title", "author", "node")

    def __init__(self, player: Player, track: Track, node: Node):
        super().__init__(player, track, node)
        self.url = track.uri
        self.identifier = track.identifier
        self.duration = track.duration
        self.title = track.title
        self.author = track.author


class TrackStartTikTokEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a TikTok track.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    url: :class:`str`
        The url of the track that started playing.
    identifier: :class:`str`
        The identifier of the track that started playing.
    duration: :class:`int`
        The duration of the track that started playing.
    title: :class:`str`
        The title of the track that started playing.
    author: :class:`str`
        The author of the track that started playing.
    node: :class:`Node`
        The node that the track started playing on.

    Parameters
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    node: :class:`Node`
        The node that the track started playing on.
    """

    __slots__ = ("player", "track", "url", "identifier", "duration", "title", "author", "node")

    def __init__(self, player: Player, track: Track, node: Node):
        super().__init__(player, track, node)
        self.url = track.uri
        self.identifier = track.identifier
        self.duration = track.duration
        self.title = track.title
        self.author = track.author


class TrackStartSpotifyEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Spotify track.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    url: :class:`str`
        The url of the track that started playing.
    identifier: :class:`str`
        The identifier of the track that started playing.
    duration: :class:`int`
        The duration of the track that started playing.
    title: :class:`str`
        The title of the track that started playing.
    author: :class:`str`
        The author of the track that started playing.
    node: :class:`Node`
        The node that the track started playing on.

    Parameters
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    node: :class:`Node`
        The node that the track started playing on.
    """

    __slots__ = ("player", "track", "url", "identifier", "duration", "title", "author", "node")

    def __init__(self, player: Player, track: Track, node: Node):
        super().__init__(player, track, node)
        self.url = track.uri
        self.identifier = track.identifier
        self.duration = track.duration
        self.title = track.title
        self.author = track.author


class TrackStartAppleMusicEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Apple Music track.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    url: :class:`str`
        The url of the track that started playing.
    identifier: :class:`str`
        The identifier of the track that started playing.
    duration: :class:`int`
        The duration of the track that started playing.
    title: :class:`str`
        The title of the track that started playing.
    author: :class:`str`
        The author of the track that started playing.
    node: :class:`Node`
        The node that the track started playing on.

    Parameters
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    node: :class:`Node`
        The node that the track started playing on.
    """

    __slots__ = ("player", "track", "url", "identifier", "duration", "title", "author", "node")

    def __init__(self, player: Player, track: Track, node: Node):
        super().__init__(player, track, node)
        self.url = track.uri
        self.identifier = track.identifier
        self.duration = track.duration
        self.title = track.title
        self.author = track.author


class TrackStartBandcampEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Bandcamp track.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    url: :class:`str`
        The url of the track that started playing.
    identifier: :class:`str`
        The identifier of the track that started playing.
    duration: :class:`int`
        The duration of the track that started playing.
    title: :class:`str`
        The title of the track that started playing.
    author: :class:`str`
        The author of the track that started playing.
    node: :class:`Node`
        The node that the track started playing on.

    Parameters
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    node: :class:`Node`
        The node that the track started playing on.
    """

    __slots__ = ("player", "track", "url", "identifier", "duration", "title", "author", "node")

    def __init__(self, player: Player, track: Track, node: Node):
        super().__init__(player, track, node)
        self.url = track.uri
        self.identifier = track.identifier
        self.duration = track.duration
        self.title = track.title
        self.author = track.author


class TrackStartYouTubeMusicEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a YouTube Music track.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    url: :class:`str`
        The url of the track that started playing.
    identifier: :class:`str`
        The identifier of the track that started playing.
    duration: :class:`int`
        The duration of the track that started playing.
    title: :class:`str`
        The title of the track that started playing.
    author: :class:`str`
        The author of the track that started playing.
    node: :class:`Node`
        The node that the track started playing on.

    Parameters
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    node: :class:`Node`
        The node that the track started playing on.
    """

    __slots__ = ("player", "track", "url", "identifier", "duration", "title", "author", "node")

    def __init__(self, player: Player, track: Track, node: Node):
        super().__init__(player, track, node)
        self.url = track.uri
        self.identifier = track.identifier
        self.duration = track.duration
        self.title = track.title
        self.author = track.author


class TrackStartSoundCloudEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a SoundCloud track.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    url: :class:`str`
        The url of the track that started playing.
    identifier: :class:`str`
        The identifier of the track that started playing.
    duration: :class:`int`
        The duration of the track that started playing.
    title: :class:`str`
        The title of the track that started playing.
    author: :class:`str`
        The author of the track that started playing.
    node: :class:`Node`
        The node that the track started playing on.

    Parameters
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    node: :class:`Node`
        The node that the track started playing on.
    """

    __slots__ = ("player", "track", "url", "identifier", "duration", "title", "author", "node")

    def __init__(self, player: Player, track: Track, node: Node):
        super().__init__(player, track, node)
        self.url = track.uri
        self.identifier = track.identifier
        self.duration = track.duration
        self.title = track.title
        self.author = track.author


class TrackStartTwitchEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Twitch track.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    url: :class:`str`
        The url of the track that started playing.
    identifier: :class:`str`
        The identifier of the track that started playing.
    duration: :class:`int`
        The duration of the track that started playing.
    title: :class:`str`
        The title of the track that started playing.
    author: :class:`str`
        The author of the track that started playing.
    node: :class:`Node`
        The node that the track started playing on.

    Parameters
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    node: :class:`Node`
        The node that the track started playing on.
    """

    __slots__ = ("player", "track", "url", "identifier", "duration", "title", "author", "node")

    def __init__(self, player: Player, track: Track, node: Node):
        super().__init__(player, track, node)
        self.url = track.uri
        self.identifier = track.identifier
        self.duration = track.duration
        self.title = track.title
        self.author = track.author


class TrackStartHTTPEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a HTTP track.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    url: :class:`str`
        The url of the track that started playing.
    identifier: :class:`str`
        The identifier of the track that started playing.
    duration: :class:`int`
        The duration of the track that started playing.
    title: :class:`str`
        The title of the track that started playing.
    author: :class:`str`
        The author of the track that started playing.
    node: :class:`Node`
        The node that the track started playing on.

    Parameters
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    node: :class:`Node`
        The node that the track started playing on.
    """

    __slots__ = ("player", "track", "url", "identifier", "duration", "title", "author", "node")

    def __init__(self, player: Player, track: Track, node: Node):
        super().__init__(player, track, node)
        self.url = track.uri
        self.identifier = track.identifier
        self.duration = track.duration
        self.title = track.title
        self.author = track.author


class TrackStartLocalFileEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a local file track.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    url: :class:`str`
        The url of the track that started playing.
    identifier: :class:`str`
        The identifier of the track that started playing.
    duration: :class:`int`
        The duration of the track that started playing.
    title: :class:`str`
        The title of the track that started playing.
    author: :class:`str`
        The author of the track that started playing.
    node: :class:`Node`
        The node that the track started playing on.

    Parameters
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    node: :class:`Node`
        The node that the track started playing on.
    """

    __slots__ = ("player", "track", "url", "identifier", "duration", "title", "author", "node")

    def __init__(self, player: Player, track: Track, node: Node):
        super().__init__(player, track, node)
        self.url = track.uri
        self.identifier = track.identifier
        self.duration = track.duration
        self.title = track.title
        self.author = track.author


class TrackStartNicoNicoEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a NicoNico track.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    url: :class:`str`
        The url of the track that started playing.
    identifier: :class:`str`
        The identifier of the track that started playing.
    duration: :class:`int`
        The duration of the track that started playing.
    title: :class:`str`
        The title of the track that started playing.
    author: :class:`str`
        The author of the track that started playing.
    node: :class:`Node`
        The node that the track started playing on.

    Parameters
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    node: :class:`Node`
        The node that the track started playing on.
    """

    __slots__ = ("player", "track", "url", "identifier", "duration", "title", "author", "node")

    def __init__(self, player: Player, track: Track, node: Node):
        super().__init__(player, track, node)
        self.url = track.uri
        self.identifier = track.identifier
        self.duration = track.duration
        self.title = track.title
        self.author = track.author


class TrackStartVimeoEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Vimeo track.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    url: :class:`str`
        The url of the track that started playing.
    identifier: :class:`str`
        The identifier of the track that started playing.
    duration: :class:`int`
        The duration of the track that started playing.
    title: :class:`str`
        The title of the track that started playing.
    author: :class:`str`
        The author of the track that started playing.
    node: :class:`Node`
        The node that the track started playing on.

    Parameters
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    node: :class:`Node`
        The node that the track started playing on.
    """

    __slots__ = ("player", "track", "url", "identifier", "duration", "title", "author", "node")

    def __init__(self, player: Player, track: Track, node: Node):
        super().__init__(player, track, node)
        self.url = track.uri
        self.identifier = track.identifier
        self.duration = track.duration
        self.title = track.title
        self.author = track.author


class TrackStartSpeakEvent(TrackStartEvent):
    """This event is dispatched when t he player starts to play a Speak track.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    url: :class:`str`
        The url of the track that started playing.
    identifier: :class:`str`
        The identifier of the track that started playing.
    duration: :class:`int`
        The duration of the track that started playing.
    title: :class:`str`
        The title of the track that started playing.
    author: :class:`str`
        The author of the track that started playing.
    node: :class:`Node`
        The node that the track started playing on.

    Parameters
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    node: :class:`Node`
        The node that the track started playing on.
    """

    __slots__ = ("player", "track", "url", "identifier", "duration", "title", "author", "node")

    def __init__(self, player: Player, track: Track, node: Node):
        super().__init__(player, track, node)
        self.url = track.uri
        self.identifier = track.identifier
        self.duration = track.duration
        self.title = track.title
        self.author = track.author


class TrackStartGCTTSEvent(TrackStartEvent):
    """This event is dispatched when the player starts to play a Google Cloud TTS track.

    Attributes
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    url: :class:`str`
        The url of the track that started playing.
    identifier: :class:`str`
        The identifier of the track that started playing.
    duration: :class:`int`
        The duration of the track that started playing.
    title: :class:`str`
        The title of the track that started playing.
    author: :class:`str`
        The author of the track that started playing.
    node: :class:`Node`
        The node that the track started playing on.

    Parameters
    ----------
    player: :class:`Player`
        The player that started to play a track.
    track: :class:`Track`
        The track that started playing.
    node: :class:`Node`
        The node that the track started playing on.
    """

    __slots__ = ("player", "track", "url", "identifier", "duration", "title", "author", "node")

    def __init__(self, player: Player, track: Track, node: Node):
        super().__init__(player, track, node)
        self.url = track.uri
        self.identifier = track.identifier
        self.duration = track.duration
        self.title = track.title
        self.author = track.author


class PlayerUpdateEvent(Event):
    """This event is dispatched when the player's progress changes.

    Attributes
    ----------
    player: :class:`Player`
        The player that's progress was updated.
    position: :class:`int`
        The position of the player that was changed to.
    timestamp: :class:`int`
        The timestamp that the player is currently on.

    Parameters
    ----------
    player: :class:`Player`
        The player that's progress was updated.
    position: :class:`int`
        The position of the player that was changed to.
    timestamp: :class:`int`
        The timestamp that the player is currently on.
    """

    __slots__ = ("player", "position", "timestamp")

    def __init__(self, player: Player, position: float, timestamp: float):
        self.player = player
        self.position = position
        self.timestamp = timestamp


class NodeDisconnectedEvent(Event):
    """This event is dispatched when a node disconnects and becomes unavailable.

    Attributes
    ----------
    node: :class:`Node`
        The node that was disconnected from.
    code: :class:`int`
        The status code of the event.
    reason: :class:`str`
        The reason of why the node was disconnected.

    Parameters
    ----------
    node: :class:`Node`
        The node that was disconnected from.
    code: :class:`int`
        The status code of the event.
    reason: :class:`str`
        The reason of why the node was disconnected.
    """

    __slots__ = ("node", "code", "reason")

    def __init__(self, node: Node, code: int, reason: str):
        self.node = node
        self.code = code
        self.reason = reason


class NodeConnectedEvent(Event):
    """This event is dispatched when Lavalink.py successfully connects to a node.

    Attributes
    ----------
    node: :class:`Node`
        The node that was successfully connected to.

    Parameters
    ----------
    node: :class:`Node`
        The node that was successfully connected to.
    """

    __slots__ = ("node",)

    def __init__(self, node: Node):
        self.node = node


class NodeChangedEvent(Event):
    """This event is dispatched when a player changes to another node.
    Keep in mind this event can be dispatched multiple times if a node
    disconnects and the load balancer moves players to a new node.


    Attributes
    ----------
    player: :class:`Player`
        The player whose node was changed.
    old_node: :class:`Node`
        The node the player was moved from.
    new_node: :class:`Node`
        The node the player was moved to.

    Parameters
    ----------
    player: :class:`Player`
        The player whose node was changed.
    old_node: :class:`Node`
        The node the player was moved from.
    new_node: :class:`Node`
        The node the player was moved to.
    """

    __slots__ = ("player", "old_node", "new_node")

    def __init__(self, player: Player, old_node: Node, new_node: Node):
        self.player = player
        self.old_node = old_node
        self.new_node = new_node


class WebSocketClosedEvent(Event):
    """This event is dispatched when an audio websocket to Discord
    is closed. This can happen for various reasons like an
    expired voice server update.

    Attributes
    ----------
    player: :class:`Player`
        The player whose audio websocket was closed.
    code: :class:`int`
        The node the player was moved from.
    reason: :class:`str`
        The node the player was moved to.
    by_remote: :class:`bool`
        If the websocket was closed remotely.
    channel: :class:`discord.channel.VocalGuildChannel`
        The voice channel the player was in.
    node: :class:`Node`
        The node the player was in.

    Parameters
    ----------
    player: :class:`Player`
        The player whose audio websocket was closed.
    code: :class:`int`
        The node the player was moved from.
    reason: :class:`str`
        The node the player was moved to.
    by_remote: :class:`bool`
        If the websocket was closed remotely.
    channel: :class:`discord.channel.VocalGuildChannel`
        The voice channel the player was in.
    node: :class:`Node`
        The node the player was in.
    """

    __slots__ = ("player", "code", "reason", "by_remote", "node", "channel")

    def __init__(
        self,
        player: Player,
        code: int,
        reason: str,
        by_remote: bool,
        node: Node,
        channel: discord.channel.VocalGuildChannel,
    ):
        self.player = player
        self.code = code
        self.reason = reason
        self.by_remote = by_remote
        self.node = node
        self.channel = channel


class SegmentSkippedEvent(Event):
    """This event is dispatched when a segment is skipped.

    Attributes
    ----------
    player: :class:`Player`
        The player whose segment was skipped.
    segment: :class:`Segment`
        The segment that was skipped.
    node: :class:`Node`
        The node the player was in.

    Parameters
    ----------
    player: :class:`Player`
        The player whose segment was skipped.
    segment: :class:`Segment`
        The segment that was skipped.
    node: :class:`Node`
        The node the player was in.
    """

    __slots__ = ("player", "category", "start", "end", "node")

    def __init__(self, player: Player, category: str, start: float, end: float, node: Node):
        self.player = player
        self.segment = Segment(category=category, start=start, end=end)
        self.node = node


class SegmentsLoadedEvent(Event):
    """This event is dispatched when segments are loaded.

    Attributes
    ----------
    player: :class:`Player`
        The player whose segments were loaded.
    segments: :class:`list` of :class:`Segment`
        The segments that were loaded.
    node: :class:`Node`
        The node the player was in.

    Parameters
    ----------
    player: :class:`Player`
        The player whose segments were loaded.
    segments: :class:`list` of :class:`Segment`
        The segments that were loaded.
    node: :class:`Node`
        The node the player was in.
    """

    __slots__ = ("player", "segments", "node")

    def __init__(self, player: Player, segments: list[dict], node: Node):
        self.player = player
        self.segments = [Segment(**segment) for segment in segments]
        self.node = node


class PlayerPausedEvent(Event):
    """This event is dispatched when a player is paused.

    Attributes
    ----------
    player: :class:`Player`
        The player whose track was paused.
    requester: :class:`discord.Member`
        The user who requested the change.

    Parameters
    ----------
    player: :class:`Player`
        The player whose track was paused.
    requester: :class:`discord.Member`
        The user who requested the change.
    """

    __slots__ = ("player", "requester")

    def __init__(self, player: Player, requester: discord.Member):
        self.player = player
        self.user = requester


class PlayerStoppedEvent(Event):
    """This event is dispatched when a player is stopped.

    Attributes
    ----------
    player: :class:`Player`
        The player which was stopped.
    requester: :class:`discord.Member`
        The user who requested the change.

    Parameters
    ----------
    player: :class:`Player`
        The player which was stopped.
    requester: :class:`discord.Member`
        The user who requested the change.
    """

    __slots__ = ("player", "requester")

    def __init__(self, player: Player, requester: discord.Member):
        self.player = player
        self.requester = requester


class PlayerResumedEvent(Event):
    """This event is dispatched when a player is resumed.

    Attributes
    ----------
    player: :class:`Player`
        The player whose track was resumed.
    requester: :class:`discord.Member`
        The user who requested the change.

    Parameters
    ----------
    player: :class:`Player`
        The player whose track was resumed.
    requester: :class:`discord.Member`
        The user who requested the change.
    """

    __slots__ = ("player", "requester")

    def __init__(self, player: Player, requester: discord.Member):
        self.player = player
        self.requester = requester


class PlayerRestoredEvent(Event):
    """This event is dispatched when a player is restored.

    Attributes
    ----------
    player: :class:`Player`
        The player whose track was restored.
    requester: :class:`discord.Member`
        The user who requested the change.

    Parameters
    ----------
    player: :class:`Player`
        The player whose track was restored.
    requester: :class:`discord.Member`
        The user who requested the change.
    """

    __slots__ = ("player", "requester")

    def __init__(self, player: Player, requester: discord.abc.User):
        self.player = player
        self.requester = requester


class QueueTrackPositionChangedEvent(Event):
    """This event is dispatched when the position of a track is changed.

    Attributes
    ----------
    player: :class:`Player`
        The player whose track position was changed.
    before: :class:`int`
        The position of the track before the change.
    after: :class:`int`
        The position of the track after the change.
    requester: :class:`discord.Member`
        The user who requested the change.
    track: :class:`Track`
        The track whose position was changed.

    Parameters
    ----------
    player: :class:`Player`
        The player whose track position was changed.
    before: :class:`int`
        The position of the track before the change.
    after: :class:`int`
        The position of the track after the change.
    requester: :class:`discord.Member`
        The user who requested the change.
    track: :class:`Track`
        The track whose position was changed.
    """

    __slots__ = ("player", "before", "after", "requester", "track")

    def __init__(self, player: Player, before: int, after: int, requester: discord.Member, track: Track):
        self.player = player
        self.requester = requester
        self.before = before
        self.after = after
        self.track = track


class TrackSkippedEvent(Event):
    """This event is dispatched when a track is skipped.

    Attributes
    ----------
    player: :class:`Player`
        The player whose track was skipped.
    track: :class:`Track`
        The track that was skipped.
    requester: :class:`discord.Member`
        The user who requested the change.
    position: :class:`float`
        The position of the track before the skip.

    Parameters
    ----------
    player: :class:`Player`
        The player whose track was skipped.
    track: :class:`Track`
        The track that was skipped.
    requester: :class:`discord.Member`
        The user who requested the change.
    position: :class:`float`
        The position of the track before the skip.
    """

    __slots__ = ("player", "track", "requester", "position")

    def __init__(self, player: Player, requester: discord.Member, track: Track, position: float):
        self.player = player
        self.track = track
        self.requester = requester
        self.position = position


class QueueShuffledEvent(Event):
    """This event is dispatched when the queue is shuffled.

    Attributes
    ----------
    player: :class:`Player`
        The player whose queue was shuffled.
    requester: :class:`discord.Member`
        The user who requested the change.

    Parameters
    ----------
    player: :class:`Player`
        The player whose queue was shuffled.
    requester: :class:`discord.Member`
        The user who requested the change.
    """

    __slots__ = ("player", "requester")

    def __init__(self, player: Player, requester: discord.Member):
        self.player = player
        self.requester = requester


class QueueTracksRemovedEvent(Event):
    """This event is dispatched when tracks are removed from the queue.

    Attributes
    ----------
    player: :class:`Player`
        The player whose queue was shuffled.
    requester: :class:`discord.Member`
        The user who requested the change.
    tracks: :class:`list` of :class:`Track`
        The tracks that were removed from the queue.

    Parameters
    ----------
    player: :class:`Player`
        The player whose queue was shuffled.
    requester: :class:`discord.Member`
        The user who requested the change.
    tracks: :class:`list` of :class:`Track`
        The tracks that were removed from the queue.
    """

    __slots__ = ("player", "requester", "tracks")

    def __init__(self, player: Player, requester: discord.Member, tracks: list[Track]):
        self.player = player
        self.requester = requester
        self.tracks = tracks


class FiltersAppliedEvent(Event):
    """This event is dispatched when filters are applied.

    Attributes
    ----------
    player: :class:`Player`
        The player whose queue was shuffled.
    requester: :class:`discord.Member`
        The user who requested the change.
    volume: :class:`Volume`
        The volume that was set.
    equalizer: :class:`Equalizer`
        The equalizer that was set.
    karaoke: :class:`Karaoke`
        The karaoke that was set.
    timescale: :class:`Timescale`
        The timescale that was set.
    tremolo: :class:`Tremolo`
        The tremolo that was set.
    vibrato: :class:`Vibrato`
        The vibrato that was set.
    rotation: :class:`Rotation`
        The rotation that was set.
    distortion: :class:`Distortion`
        The distortion that was set.
    low_pass: :class:`Lowpass`
        The lowpass that was set.
    channel_mix: :class:`ChannelMix`
        The channel mix that was set.
    node: :class:`Node`
        The node that was changed.

    Parameters
    ----------
    player: :class:`Player`
        The player whose queue was shuffled.
    requester: :class:`discord.Member`
        The user who requested the change.
    volume: :class:`Volume`
        The volume that was set.
    equalizer: :class:`Equalizer`
        The equalizer that was set.
    karaoke: :class:`Karaoke`
        The karaoke that was set.
    timescale: :class:`Timescale`
        The timescale that was set.
    tremolo: :class:`Tremolo`
        The tremolo that was set.
    vibrato: :class:`Vibrato`
        The vibrato that was set.
    rotation: :class:`Rotation`
        The rotation that was set.
    distortion: :class:`Distortion`
        The distortion that was set.
    low_pass: :class:`Lowpass`
        The lowpass that was set.
    channel_mix: :class:`ChannelMix`
        The channel mix that was set.
    node: :class:`Node`
        The node that was changed.
    """

    __slots__ = (
        "player",
        "requester",
        "volume",
        "equalizer",
        "karaoke",
        "timescale",
        "tremolo",
        "vibrato",
        "rotation",
        "distortion",
        "low_pass",
        "channel_mix",
        "node",
    )

    def __init__(
        self,
        player: Player,
        requester: discord.Member,
        node: Node,
        volume: Volume = None,
        equalizer: Equalizer = None,
        karaoke: Karaoke = None,
        timescale: Timescale = None,
        tremolo: Tremolo = None,
        vibrato: Vibrato = None,
        rotation: Rotation = None,
        distortion: Distortion = None,
        low_pass: LowPass = None,
        channel_mix: ChannelMix = None,
    ):
        self.player = player
        self.requester = requester
        self.volume = volume
        self.equalizer = equalizer
        self.karaoke = karaoke
        self.timescale = timescale
        self.tremolo = tremolo
        self.vibrato = vibrato
        self.rotation = rotation
        self.distortion = distortion
        self.low_pass = low_pass
        self.channel_mix = channel_mix
        self.node = node


class PlayerMovedEvent(Event):
    """This event is dispatched when the player is moved.

    Attributes
    ----------
    player: :class:`Player`
        The player whose queue was shuffled.
    requester: :class:`discord.Member`
        The user who requested the change.
    before: :class:`discord.channel.VocalGuildChannel`
        The channel the player was in before the move.
    after: :class:`discord.channel.VocalGuildChannel`
        The channel the player is in now.
    """

    __slots__ = ("player", "requester", "before", "after")

    def __init__(
        self,
        player: Player,
        requester: discord.Member,
        before: discord.channel.VocalGuildChannel,
        after: discord.channel.VocalGuildChannel,
    ):
        self.player = player
        self.requester = requester
        self.before = before
        self.after = after


class PlayerDisconnectedEvent(Event):
    """This event is dispatched when the player is moved.

    Attributes
    ----------
    player: :class:`Player`
        The player whose queue was shuffled.
    requester: :class:`discord.Member`
        The user who requested the change.
    current_track: :class:`Track`
        The track that was playing when the player disconnected.
    position: :class:`int`
        The position of the track that was playing when the player disconnected.
    queue: :class:`collections.deque` of :class:`Track`
        The tracks that are in the queue when the player disconnected.

    Parameters
    ----------
    player: :class:`Player`
        The player whose queue was shuffled.
    requester: :class:`discord.Member`
        The user who requested the change.
    """

    __slots__ = ("player", "requester", "current_track", "position", "queue")

    def __init__(self, player: Player, requester: discord.Member):
        self.player = player
        self.requester = requester
        self.current_track: Track | None = player.current
        self.position: float = player.position
        self.queue: collections.deque = player.queue.raw_queue


class PlayerConnectedEvent(Event):
    """This event is dispatched when the player is moved.

    Attributes
    ----------
    player: :class:`Player`
        The player whose queue was shuffled.
    requester: :class:`discord.Member`
        The user who requested the change.
    """

    __slots__ = ("player", "requester")

    def __init__(self, player: Player, requester: discord.Member | None):
        self.player = player
        self.requester = requester


class TrackSeekEvent(Event):
    """This event is dispatched when the player is moved.

    Attributes
    ----------
    player: :class:`Player`
        The player whose queue was shuffled.
    requester: :class:`discord.Member`
        The user who requested the change.
    track: :class:`Track`
        The track that was seeked.
    after: float
        The position in the track that was seeked to.
    before: float
        The position in the track that was seeked from.

    """

    __slots__ = ("player", "requester", "track", "before", "after")

    def __init__(self, player: Player, requester: discord.Member, track: Track, before: float, after: float):
        self.player = player
        self.requester = requester
        self.track = track
        self.before = before
        self.after = after


class PlayerVolumeChangedEvent(Event):
    """This event is dispatched when the player is moved.

    Attributes
    ----------
    player: :class:`Player`
        The player whose queue was shuffled.
    requester: :class:`discord.Member`
        The user who requested the change.
    before: float
        The volume before the change.
    after: float
        The volume after the change.

    """

    __slots__ = ("player", "requester", "before", "after")

    def __init__(self, player: Player, requester: discord.Member, before: int, after: int):
        self.player = player
        self.requester = requester
        self.before = before
        self.after = after


class PlayerRepeatEvent(Event):
    """This event is dispatched when the player is moved.

    Attributes
    ----------
    player: :class:`Player`
        The player whose queue was shuffled.
    requester: :class:`discord.Member`
        The user who requested the change.
    queue_before: bool
        The repeat state before the change.
    queue_after: bool
        The repeat state after the change.
    current_before: bool
        The repeat state before the change.
    current_after: bool
        The repeat state after the change.
    type : str
        The type of repeat that was set.

    """

    __slots__ = ("player", "requester", "queue_before", "queue_after", "current_before", "current_after", "type")

    def __init__(
        self,
        player: Player,
        requester: discord.Member,
        op_type: Literal["current", "queue"],
        queue_before: bool,
        queue_after: bool,
        current_before: bool,
        current_after: bool,
    ):
        self.player = player
        self.requester = requester
        self.queue_before = queue_before
        self.queue_after = queue_after
        self.current_before = current_before
        self.current_after = current_after
        self.type = op_type


class TrackPreviousRequestedEvent(Event):
    """This event is dispatched when a history track is requested.

    Attributes
    ----------
    player: :class:`Player`
        The player whose queue was shuffled.
    requester: :class:`discord.Member`
        The user who requested the change.
    track: :class:`Track`
        The track that was seeked.
    """

    __slots__ = ("player", "requester", "track")

    def __init__(self, player: Player, requester: discord.Member, track: Track):
        self.player = player
        self.requester = requester
        self.track = track


class TracksRequestedEvent(Event):
    """This event is dispatched when a track in added to the queue.

    Attributes
    ----------
    player: :class:`Player`
        The player whose queue was shuffled.
    requester: :class:`discord.Member`
        The user who requested the change.
    tracks: :class:`list` of :class:`Track`
        The tracks that were added to the queue.
    """

    __slots__ = ("player", "requester", "tracks")

    def __init__(self, player: Player, requester: discord.Member, tracks: list[Track]):
        self.player = player
        self.requester = requester
        self.tracks = tracks


class QuickPlayEvent(Event):
    """This event is dispatched when a track is played with higher priority than the current track.

    Attributes
    ----------
    player: :class:`Player`
        The player whose queue was shuffled.
    requester: :class:`discord.Member`
        The user who requested the change.
    track: :class:`Track`
        The track that was played.
    """

    __slots__ = ("player", "requester", "track")

    def __init__(self, player: Player, requester: discord.Member, tracks: Track):
        self.player = player
        self.requester = requester
        self.tracks = tracks



from __future__ import annotations

import asyncio
import collections
import contextlib
import dataclasses
import datetime
import functools
import random
import threading
import time
from asyncio import QueueFull, events, locks
from collections.abc import AsyncIterable, AsyncIterator, Awaitable, Coroutine, Generator, Iterable, Iterator
from copy import copy
from enum import Enum
from functools import _make_key  # type: ignore
from itertools import chain
from types import GenericAlias
from typing import TYPE_CHECKING, Any, Callable, Optional, Tuple, TypeVar, Union

import discord  # type: ignore
from async_lru import (
    _cache_clear,
    _cache_hit,
    _cache_info,
    _cache_invalidate,
    _cache_miss,
    _close,
    _done_callback,
    _open,
    unpartial,
)
from discord.backoff import ExponentialBackoff
from discord.ext import commands as dpy_command
from discord.ext.commands.view import StringView
from discord.types.embed import EmbedType
from discord.utils import MISSING as D_MISSING  # noqa
from discord.utils import maybe_coroutine

from pylav.types import BotT, CogT, ContextT, InteractionT

try:
    from redbot.core.commands import Command
    from redbot.core.commands import Context as _OriginalContextClass
except ImportError:
    from discord.ext.commands import Command
    from discord.ext.commands import Context as _OriginalContextClass

from discord.ext.commands import Context as DpyContext

if TYPE_CHECKING:
    from pylav import Player
    from pylav.node import Node

__all__ = (
    "MISSING",
    "AsyncIter",
    "add_property",
    "format_time",
    "get_time_string",
    "PlayerQueue",
    "TrackHistoryQueue",
    "SegmentCategory",
    "Segment",
    "_process_commands",
    "_get_context",
    "PyLavContext",
    "ExponentialBackoffWithReset",
)

from pylav._logging import getLogger

T = TypeVar("T")

LOGGER = getLogger("PyLav.utils")
_RED_LOGGER = getLogger("red")

_LOCK = threading.Lock()


def alru_cache(fn=None, maxsize=128, typed=False, *, cache_exceptions=True, ignore_kwargs=None):
    def wrapper(fn):
        _origin = unpartial(fn)

        if not asyncio.iscoroutinefunction(_origin):
            raise RuntimeError(f"Coroutine function is required, got {fn}")

        # functools.partialmethod support
        if hasattr(fn, "_make_unbound_method"):
            fn = fn._make_unbound_method()

        @functools.wraps(fn)
        async def wrapped(*fn_args, **fn_kwargs):
            if wrapped.closed:
                raise RuntimeError(f"alru_cache is closed for {wrapped}")

            loop = asyncio.get_event_loop()

            key = _make_key(fn_args, fn_kwargs, typed)

            fut = wrapped._cache.get(key)

            if fut is not None:
                if not fut.done():
                    _cache_hit(wrapped, key)
                    return await asyncio.shield(fut)

                exc = fut._exception

                if exc is None or cache_exceptions:
                    _cache_hit(wrapped, key)
                    return fut.result()

                # exception here and cache_exceptions == False
                wrapped._cache.pop(key)

            fut = loop.create_future()
            task = loop.create_task(fn(*fn_args, **fn_kwargs))
            task.add_done_callback(functools.partial(_done_callback, fut))
            wrapped.tasks.add(task)
            task.add_done_callback(wrapped.tasks.remove)
            wrapped._cache[key] = fut
            if maxsize is not None and len(wrapped._cache) > maxsize:
                wrapped._cache.popitem()

            _cache_miss(wrapped, key)
            return await asyncio.shield(fut)

        _cache_clear(wrapped)
        wrapped._origin = _origin
        wrapped.closed = False
        wrapped.cache_info = functools.partial(_cache_info, wrapped, maxsize)
        wrapped.cache_clear = functools.partial(_cache_clear, wrapped)
        wrapped.invalidate = functools.partial(_cache_invalidate, wrapped, typed)
        wrapped.close = functools.partial(_close, wrapped)
        wrapped.open = functools.partial(_open, wrapped)

        return wrapped

    if fn is None:
        return wrapper

    if callable(fn) or hasattr(fn, "_make_unbound_method"):
        return wrapper(fn)

    raise NotImplementedError(f"{fn} decorating is not supported")


def _synchronized(lock):
    """Synchronization decorator"""

    def wrapper(f):
        @functools.wraps(f)
        def inner_wrapper(*args, **kw):
            with lock:
                return f(*args, **kw)

        return inner_wrapper

    return wrapper


class SingletonMethods:
    _has_run = {}
    _responses = {}

    @classmethod
    @_synchronized(_LOCK)
    def reset(cls):
        cls._has_run = {}
        cls._responses = {}

    @classmethod
    @_synchronized(_LOCK)
    def run_once(cls, f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            if not cls._has_run.get(f, False):
                cls._has_run[f] = True
                output = f(*args, **kwargs)
                cls._responses[f] = output
                return output
            else:
                return cls._responses.get(f, None)

        cls._has_run[f] = False
        return wrapper

    @classmethod
    @_synchronized(_LOCK)
    def run_once_async(cls, f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            if not cls._has_run.get(f, False):
                cls._has_run[f.__name__] = True
                return f(*args, **kwargs)
            else:
                return asyncio.sleep(0)

        cls._has_run[f] = False
        return wrapper


class _Singleton(type):
    _instances = {}

    def __call__(self, *args, **kwargs):
        if self not in self._instances:
            self._locked_call(*args, **kwargs)
        return self._instances[self]

    @_synchronized(_LOCK)
    def _locked_call(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)


class MissingSentinel(str):
    def __str__(self) -> str:
        return "MISSING"

    def __repr__(self) -> str:
        return "MISSING"

    def __bool__(self) -> bool:
        return False


MISSING: Any = MissingSentinel("MISSING")


@dataclasses.dataclass(eq=True)
class TimedFeature:
    enabled: bool = False
    time: int = 60

    def to_dict(self) -> dict:
        return {"enabled": self.enabled, "time": self.time}


def add_property(inst: object, name: str, method: Callable) -> None:
    cls = type(inst)
    if not hasattr(cls, "__perinstance"):
        cls = type(cls.__name__, (cls,), {})
        cls.__perinstance = True
        inst.__class__ = cls
    setattr(cls, name, property(method))


def get_time_string(seconds: int | float) -> str:
    seconds = int(seconds)
    m, s = divmod(seconds, 60)
    h, m = divmod(m, 60)
    d, h = divmod(h, 24)

    if d > 0:
        msg = "{0}d {1}h"
    elif d == 0 and h > 0:
        msg = "{1}h {2}m"
    elif d == 0 and h == 0 and m > 0:
        msg = "{2}m {3}s"
    elif d == 0 and h == 0 and m == 0 and s > 0:
        msg = "{3}s"
    else:
        msg = ""
    return msg.format(d, h, m, s)


def format_time(duration: int | float) -> str:
    """Formats the given time into DD:HH:MM:SS"""
    seconds = int(duration // 1000)
    days, seconds = divmod(seconds, 24 * 60 * 60)
    hours, seconds = divmod(seconds, 60 * 60)
    minutes, seconds = divmod(seconds, 60)
    day = f"{days:02d}:" if days else ""
    hour = f"{hours:02d}:" if hours or day else ""
    minutes = f"{minutes:02d}:"
    sec = f"{seconds:02d}"
    return f"{day}{hour}{minutes}{sec}"


class PlayerQueue(asyncio.Queue[T]):
    """A queue, useful for coordinating producer and consumer coroutines.

    If maxsize is less than or equal to zero, the queue size is infinite. If it
    is an integer greater than 0, then "await put()" will block when the
    queue reaches maxsize, until an item is removed by get().

    Unlike the standard library Queue, you can reliably know this Queue's size
    with qsize(), since your single-threaded asyncio application won't be
    interrupted between calling qsize() and doing an operation on the Queue.
    """

    _queue: collections.deque[T]
    raw_b64s: list[str]

    def __init__(self, maxsize=0):
        self._maxsize = maxsize
        self._loop = events.get_event_loop()

        # Futures.
        self._getters = collections.deque()
        # Futures.
        self._putters = collections.deque()
        self._unfinished_tasks = 0
        self._finished = locks.Event()
        self._finished.set()
        self._init(maxsize)

    # These do not exist in asyncio.Queue
    @property
    def raw_queue(self):
        return self._queue.copy()

    @raw_queue.setter
    def raw_queue(self, value: collections.deque[T]):
        if not isinstance(value, collections.deque):
            raise TypeError("Queue value must be a collections.deque[Track]")
        if self._maxsize and len(value) > self._maxsize:
            raise ValueError(f"Queue value cannot be longer than maxsize: {self._maxsize}")
        self._queue = value

    def popindex(self, index: int) -> T:
        value = self._queue[index]
        del self._queue[index]
        return value

    async def remove(self, value: T, duplicates: bool = False) -> tuple[list[T], int]:
        """Removes the first occurrence of a value from the queue.

        If duplicates is True, all occurrences of the value are removed.
        Returns the number of occurrences removed.
        """
        count = 0
        removed = []
        try:
            i = self._queue.index(value)
            removed.append(self.popindex(i))
            count += 1
            if duplicates:
                with contextlib.suppress(ValueError):
                    while value in self:
                        i = self._queue.index(value)
                        removed.append(self.popindex(i))
                        count += 1
            return removed, count
        except ValueError as e:
            raise IndexError("Value not in queue") from e

    def clear(self):
        """Remove all items from the queue."""
        self._queue.clear()
        for i in self._getters:
            i.cancel()
        self._getters.clear()
        for i in self._putters:
            i.cancel()
        self._putters.clear()

    async def shuffle(self):
        """Shuffle the queue."""
        if self.empty():
            return
        await asyncio.to_thread(random.shuffle, self._queue)

    def index(self, value: T) -> int:
        """Return first index of value."""
        return self._queue.index(value)

    def __contains__(self, obj: T) -> bool:
        return obj in self._queue

    def __len__(self):
        return len(self._queue)

    def __index__(self):
        return len(self._queue)

    def __getitem__(self, key: int | slice) -> T | list[T]:
        return self.popindex(key) if isinstance(key, int) else NotImplemented

    def __length_hint__(self):
        return len(self._queue)

    # These three are overridable in subclasses.

    def _init(self, maxsize: int):
        self._queue = collections.deque(maxlen=maxsize or None)
        self.raw_b64s = []

    def _get(self, index: int = None) -> T:
        r = self.popindex(index) if index is not None else self._queue.popleft()
        if r.track:
            self.raw_b64s.remove(r.track)
        return r

    def _put(self, items: list[T], index: int = None):
        if index is not None:
            for i in items:
                if index < 0:
                    self._queue.append(i)
                    if i.track:
                        self.raw_b64s.append(i.track)
                else:
                    self._queue.insert(index, i)
                    if i.track:
                        self.raw_b64s.append(i.track)
                    index += 1
        else:
            self._queue.extend(items)
            self.raw_b64s.extend([i.track for i in items if i.track])

    # End of the overridable methods.

    @staticmethod
    def _wakeup_next(waiters):
        # Wake up the next waiter (if any) that isn't cancelled.
        while waiters:
            waiter = waiters.popleft()
            if not waiter.done():
                waiter.set_result(None)
                break

    def __repr__(self):
        return f"<{type(self).__name__} at {id(self):#x} {self._format()}>"

    def __str__(self):
        return f"<{type(self).__name__} {self._format()}>"

    __class_getitem__ = classmethod(GenericAlias)

    def _format(self):
        result = f"maxsize={self._maxsize!r}"
        if getattr(self, "_queue", None):
            result += f" _queue={list(self._queue)!r}"
        if self._getters:
            result += f" _getters[{len(self._getters)}]"
        if self._putters:
            result += f" _putters[{len(self._putters)}]"
        if self._unfinished_tasks:
            result += f" tasks={self._unfinished_tasks}"
        return result

    def qsize(self) -> int:
        """Number of items in the queue."""
        return len(self._queue)

    size = qsize

    @property
    def maxsize(self) -> int:
        """Number of items allowed in the queue."""
        return self._maxsize

    def empty(self) -> bool:
        """Return True if the queue is empty, False otherwise."""
        return not len(self._queue)

    def full(self) -> bool:
        """Return True if there are maxsize items in the queue.

        Note: if the Queue was initialized with maxsize=0 (the default),
        then full() is never True.
        """
        return False if self._maxsize <= 0 else self.qsize() >= self._maxsize

    async def put(self, items: list[T], index: int = None) -> None:
        """Put an item into the queue.

        Put an item into the queue. If the queue is full, wait until a free
        slot is available before adding item.
        """
        while self.full():
            putter = self._loop.create_future()
            self._putters.append(putter)
            try:
                await putter
            except BaseException:
                putter.cancel()  # Just in case putter is not done yet.
                with contextlib.suppress(ValueError):
                    # Clean self._putters from canceled putters.
                    self._putters.remove(putter)
                if not self.full() and not putter.cancelled():
                    # We were woken up by get_nowait(), but can't take
                    # the call.  Wake up the next in line.
                    self._wakeup_next(self._putters)
                raise
        return self.put_nowait(items, index)

    def put_nowait(self, items: list[T], index: int = None) -> None:
        """Put an item into the queue without blocking.

        If no free slot is immediately available, raise QueueFull.
        """
        if self.full():
            raise QueueFull
        self._put(items, index=index)
        self._unfinished_tasks += 1
        self._finished.clear()
        self._wakeup_next(self._getters)

    async def get(self, index: int = None) -> T:
        """Remove and return an item from the queue.

        If queue is empty, wait until an item is available.
        """
        while self.empty():
            getter = self._loop.create_future()
            self._getters.append(getter)
            try:
                await getter
            except BaseException:
                getter.cancel()
                with contextlib.suppress(ValueError):
                    self._getters.remove(getter)
                if not self.empty() and not getter.cancelled():
                    self._wakeup_next(self._getters)
                raise
        return self.get_nowait(index=index)

    def get_nowait(self, index: int = None) -> T:
        """Remove and return an item from the queue.

        Return an item if one is immediately available, else raise QueueEmpty.
        """
        if self.empty():
            return None
        item = self._get(index=index)
        self._wakeup_next(self._putters)
        return item

    def task_done(self):
        """Indicate that a formerly enqueued task is complete.

        Used by queue consumers. For each get() used to fetch a task,
        a subsequent call to task_done() tells the queue that the processing
        on the task is complete.

        If a join() is currently blocking, it will resume when all items have
        been processed (meaning that a task_done() call was received for every
        item that had been put() into the queue).

        Raises ValueError if called more times than there were items placed in
        the queue.
        """
        if self._unfinished_tasks <= 0:
            raise ValueError("task_done() called too many times")
        self._unfinished_tasks -= 1
        if self._unfinished_tasks == 0:
            self._finished.set()

    async def join(self):
        """Block until all items in the queue have been gotten and processed.

        The count of unfinished tasks goes up whenever an item is added to the
        queue. The count goes down whenever a consumer calls task_done() to
        indicate that the item was retrieved and all work on it is complete.
        When the count of unfinished tasks drops to zero, join() unblocks.
        """
        if self._unfinished_tasks > 0:
            await self._finished.wait()


class TrackHistoryQueue(PlayerQueue[T]):
    def __int__(self, maxsize: int = 0) -> None:
        super().__init__(maxsize=maxsize)

    def _init(self, maxsize: int):
        super()._init(maxsize)

    def _put(self, items: list[T], index: int = None):
        if len(items) + self.qsize() > self.maxsize:
            diff = len(items) + self.qsize() - self.maxsize
            for _ in range(diff):
                self._queue.pop()
                self.raw_b64s.pop()
        if index is not None:
            for i in items:
                if index < 0:
                    self._queue.append(i)
                    self.raw_b64s.append(i.track)
                else:
                    self._queue.insert(index, i)
                    self.raw_b64s.insert(index, i.track)
                    index += 1
        else:
            self._queue.extendleft(items)
            for i, t in enumerate(items):
                self.raw_b64s.insert(i, t.track)

    def _get(self, index: int = None) -> T:
        r = self.popindex(index) if index is not None else self._queue.popleft()
        self.raw_b64s.pop(index if index is not None else -1)
        return r

    def put_nowait(self, items: list[T], index: int = None):
        """Put an item into the queue without blocking.

        If no free slot is immediately available, raise QueueFull.
        """
        self._put(items, index=index)
        self._unfinished_tasks += 1
        self._finished.clear()
        self._wakeup_next(self._getters)

    def get_nowait(self, index: int = None) -> T | None:
        """Remove and return an item from the queue.

        Return an item if one is immediately available, else raise QueueEmpty.
        """
        if self.empty():
            return None
        item = self._get(index=index)
        self._wakeup_next(self._putters)
        return item


class SegmentCategory(Enum):
    """
    Segment category
    """

    Sponsor = "sponsor"
    Selfpromo = "selfpromo"
    Interaction = "interaction"
    Intro = "intro"
    Outro = "outro"
    Preview = "preview"
    MusicOfftopic = "music_offtopic"
    Filler = "filler"

    @classmethod
    def get_category(cls, segment_type: str) -> SegmentCategory:
        """
        Get segment category
        """
        if segment_type == "intro":
            return SegmentCategory.Intro
        elif segment_type == "outro":
            return SegmentCategory.Outro
        elif segment_type == "preview":
            return SegmentCategory.Preview
        elif segment_type == "music_offtopic":
            return SegmentCategory.MusicOfftopic
        elif segment_type == "filler":
            return SegmentCategory.Filler
        elif segment_type == "sponsor":
            return SegmentCategory.Sponsor
        elif segment_type == "selfpromo":
            return SegmentCategory.Selfpromo
        elif segment_type == "interaction":
            return SegmentCategory.Interaction
        else:
            raise ValueError(f"Unknown segment type: {segment_type}")

    @classmethod
    def get_category_name(cls, segment_type: str) -> str:
        """
        Get segment category name
        """
        return cls.get_category(segment_type).name

    @classmethod
    def get_category_from_name(cls, category_name: str) -> SegmentCategory:
        """
        Get segment category from name
        """
        return SegmentCategory[category_name]

    @classmethod
    def get_category_list(cls) -> list[SegmentCategory]:
        """
        Get segment category list
        """
        return list(cls)

    @classmethod
    def get_category_list_name(cls) -> list[str]:
        """
        Get segment category list name
        """
        return [category.name for category in cls]

    @classmethod
    def get_category_list_value(cls) -> list[str]:
        """
        Get segment category list value
        """
        return [category.value for category in cls]


class Segment:
    __slots__ = ("category", "start", "end")

    def __init__(self, /, category: str, start: float, end: float):
        self.category = category
        self.start = start
        self.end = end


class ExponentialBackoffWithReset(ExponentialBackoff):
    """
    Exponential backoff with reset
    """

    def __init__(self, base: int = 1, *, integral: T = False):
        super().__init__(base=base, integral=integral)

    def reset(self):
        """
        Reset
        """
        self._last_invocation: float = time.monotonic()
        self._exp = 0


class PyLavContext(_OriginalContextClass):
    _original_ctx_or_interaction: ContextT | InteractionT | None
    bot: BotT
    client: BotT
    interaction: InteractionT | None

    def __init__(
        self,
        *,
        message: discord.Message,
        bot: BotT,
        view: StringView,
        args: list[Any] = D_MISSING,
        kwargs: dict[str, Any] = D_MISSING,
        prefix: str | None = None,
        command: Command[Any, ..., Any] | None = None,  # noqa
        invoked_with: str | None = None,
        invoked_parents: list[str] = D_MISSING,
        invoked_subcommand: Command[Any, ..., Any] | None = None,  # noqa
        subcommand_passed: str | None = None,
        command_failed: bool = False,
        current_parameter: discord.ext.commands.Parameter | None = None,
        current_argument: str | None = None,
        interaction: InteractionT | None = None,
    ):
        super().__init__(
            message=message,
            bot=bot,
            view=view,
            args=args,
            kwargs=kwargs,
            prefix=prefix,
            command=command,
            invoked_with=invoked_with,
            invoked_parents=invoked_parents,
            invoked_subcommand=invoked_subcommand,
            subcommand_passed=subcommand_passed,
            command_failed=command_failed,
            current_parameter=current_parameter,
            current_argument=current_argument,
            interaction=interaction,
        )

        self._original_ctx_or_interaction = None
        self.lavalink = bot.lavalink
        self.pylav = bot.lavalink

    @discord.utils.cached_property
    def author(self) -> discord.User | discord.Member:
        """Union[:class:`~discord.User`, :class:`.Member`]:
        Returns the author associated with this context's command. Shorthand for :attr:`.Message.author`
        """
        # When using client.get_context() on  a button interaction the "author" becomes the bot user
        #   This ensures the original author remains the author of the context
        if isinstance(self._original_ctx_or_interaction, discord.Interaction):
            return self._original_ctx_or_interaction.user
        elif isinstance(self._original_ctx_or_interaction, DpyContext):
            return self._original_ctx_or_interaction.author
        else:
            return self.message.author

    @property
    def cog(self) -> CogT | None:
        """Optional[:class:`.Cog`]: Returns the cog associated with this context's command. None if it does not
        exist."""

        return None if self.command is None else self.command.cog

    @discord.utils.cached_property
    def guild(self) -> discord.Guild | None:
        """Optional[:class:`.Guild`]: Returns the guild associated with this context's command. None if not
        available."""
        return getattr(self.author, "guild", None)

    @discord.utils.cached_property
    def channel(self) -> discord.abc.MessageableChannel:
        """Union[:class:`.abc.Messageable`]: Returns the channel associated with this context's command.
        Shorthand for :attr:`.Message.channel`.
        """
        if isinstance(self._original_ctx_or_interaction, (discord.Interaction, DpyContext)):
            return self._original_ctx_or_interaction.channel  # type: ignore
        else:
            return self.message.channel

    @property
    def player(self) -> Player | None:
        """
        Get player
        """
        return self.lavalink.get_player(self.guild)

    async def connect_player(self, channel: discord.channel.VocalGuildChannel = None, self_deaf: bool = True) -> Player:
        """
        Connect player
        """
        requester = self.author
        channel = channel or self.author.voice.channel
        return await self.lavalink.connect_player(requester=requester, channel=channel, self_deaf=self_deaf)

    @property
    def original_ctx_or_interaction(self) -> ContextT | InteractionT | None:
        """
        Get original ctx or interaction
        """
        return self._original_ctx_or_interaction

    async def construct_embed(
        self,
        *,
        embed: discord.Embed = None,
        colour: discord.Colour | int | None = None,
        color: discord.Colour | int | None = None,
        title: str = None,
        type: EmbedType = "rich",
        url: str = None,
        description: str = None,
        timestamp: datetime.datetime = None,
        author_name: str = None,
        author_url: str = None,
        thumbnail: str = None,
        footer: str = None,
        footer_url: str = None,
        messageable: discord.abc.Messageable | InteractionT = None,
    ) -> discord.Embed:
        """
        Construct embed
        """
        return await self.lavalink.construct_embed(
            embed=embed,
            colour=colour,
            color=color,
            title=title,
            type=type,
            url=url,
            description=description,
            timestamp=timestamp,
            author_name=author_name,
            author_url=author_url,
            thumbnail=thumbnail,
            footer=footer,
            footer_url=footer_url,
            messageable=messageable or self,
        )

    @classmethod
    async def from_interaction(cls, interaction: InteractionT, /) -> PyLavContext:
        #  When using this on a button interaction it raises an error as expected.
        #   This makes the `get_context` method work with buttons by storing the original context

        added_dummy = False
        if isinstance(interaction, discord.Interaction) and interaction.command is None:
            setattr(interaction, "_cs_command", dummy_command)
            added_dummy = True
        instance = await super().from_interaction(interaction)
        if added_dummy:
            instance.command = None
        instance._original_ctx_or_interaction = interaction
        return instance

    def dispatch_command(
        self, message: discord.Message, command: Command, author: discord.abc.User, args: list[str], prefix: str = None
    ) -> None:
        """
        Dispatch command
        """
        command_str = f"{prefix}{command.qualified_name} {' '.join(args)}"
        msg = copy(message)
        msg.author = author
        msg.content = command_str
        self.bot.dispatch("message", msg)


async def _process_commands(self, message: discord.Message, /):
    """
    Same as base method, but dispatches an additional event for cogs
    which want to handle normal messages differently to command
    messages,  without the overhead of additional get_context calls
    per cog.
    """
    if not message.author.bot:
        ctx = await self.get_context(message, cls=PyLavContext)
        if ctx.invoked_with and isinstance(message.channel, discord.PartialMessageable):
            _RED_LOGGER.warning(
                "Discarded a command message (ID: %s) with PartialMessageable channel: %r",
                message.id,
                message.channel,
            )
        else:
            await self.invoke(ctx)
    else:
        ctx = None

    if ctx is None or ctx.valid is False:
        self.dispatch("message_without_command", message)


async def _get_context(self: BotT, message: discord.Message | InteractionT, /, *, cls=PyLavContext) -> PyLavContext:
    return await super(self.__class__, self).get_context(message, cls=cls)  # noqa


@dpy_command.command(name="__dummy_command", hidden=True, disabled=True)
async def dummy_command(self, context: PyLavContext):
    """Does nothing."""


async def sort_key_nodes(node: Node, region: str = None) -> float:
    return await node.penalty_with_region(region)


# Everything under here is taken from
#   https://github.com/Cog-Creators/Red-DiscordBot/blob/V3/develop/redbot/core/utils/__init__.py
#   and is licensed under the GPLv3 license (https://github.com/Cog-Creators/Red-DiscordBot/blob/V3/develop/LICENSE

_T = TypeVar("_T")
_S = TypeVar("_S")

# https://github.com/PyCQA/pylint/issues/2717


# Benchmarked to be the fastest method.
def deduplicate_iterables(*iterables):
    """
    Returns a list of all unique items in ``iterables``, in the order they
    were first encountered.
    """
    # dict insertion order is guaranteed to be preserved in 3.6+
    return list(dict.fromkeys(chain.from_iterable(iterables)))


# https://github.com/PyCQA/pylint/issues/2717
class AsyncFilter(AsyncIterator[_T], Awaitable[list[_T]]):  # pylint: disable=duplicate-bases
    """Class returned by `async_filter`. See that function for details.

    We don't recommend instantiating this class directly.
    """

    def __init__(
        self,
        func: Callable[[_T], bool | Awaitable[bool]],
        iterable: AsyncIterable[_T] | Iterable[_T],
    ) -> None:
        self.__func: Callable[[_T], bool | Awaitable[bool]] = func
        self.__iterable: AsyncIterable[_T] | Iterable[_T] = iterable

        # We assign the generator strategy based on the arguments' types
        if isinstance(iterable, AsyncIterable):
            if asyncio.iscoroutinefunction(func):
                self.__generator_instance = self.__async_generator_async_pred()
            else:
                self.__generator_instance = self.__async_generator_sync_pred()
        elif asyncio.iscoroutinefunction(func):
            self.__generator_instance = self.__sync_generator_async_pred()
        else:
            raise TypeError("Must be either an async predicate, an async iterable, or both.")

    async def __sync_generator_async_pred(self) -> AsyncIterator[_T]:
        for item in self.__iterable:
            if await self.__func(item):
                yield item

    async def __async_generator_sync_pred(self) -> AsyncIterator[_T]:
        async for item in self.__iterable:
            if self.__func(item):
                yield item

    async def __async_generator_async_pred(self) -> AsyncIterator[_T]:
        async for item in self.__iterable:
            if await self.__func(item):
                yield item

    async def __flatten(self) -> list[_T]:
        return [item async for item in self]

    def __aiter__(self):
        return self

    def __await__(self):
        # Simply return the generator filled into a list
        return self.__flatten().__await__()

    def __anext__(self) -> Awaitable[_T]:
        # This will use the generator strategy set in __init__
        return self.__generator_instance.__anext__()


def async_filter(
    func: Callable[[_T], bool | Awaitable[bool]],
    iterable: AsyncIterable[_T] | Iterable[_T],
) -> AsyncFilter[_T]:
    """Filter an (optionally async) iterable with an (optionally async) predicate.

    At least one of the arguments must be async.

    Parameters
    ----------
    func : Callable[[T], Union[bool, Awaitable[bool]]]
        A function or coroutine function which takes one item of ``iterable``
        as an argument, and returns ``True`` or ``False``.
    iterable : Union[AsyncIterable[_T], Iterable[_T]]
        An iterable or async iterable which is to be filtered.

    Raises
    ------
    TypeError
        If neither of the arguments are async.

    Returns
    -------
    AsyncFilter[T]
        An object which can either be awaited to yield a list of the filtered
        items, or can also act as an async iterator to yield items one by one.

    """
    return AsyncFilter(func, iterable)


async def async_enumerate(async_iterable: AsyncIterable[_T], start: int = 0) -> AsyncIterator[tuple[int, _T]]:
    """Async iterable version of `enumerate`.

    Parameters
    ----------
    async_iterable : AsyncIterable[T]
        The iterable to enumerate.
    start : int
        The index to start from. Defaults to 0.

    Returns
    -------
    AsyncIterator[Tuple[int, T]]
        An async iterator of tuples in the form of ``(index, item)``.

    """
    async for item in async_iterable:
        yield start, item
        start += 1


async def _sem_wrapper(sem, task):
    async with sem:
        return await task


def bounded_gather_iter(
    *coros_or_futures, limit: int = 4, semaphore: asyncio.Semaphore | None = None
) -> Iterator[Awaitable[Any]]:
    """
    An iterator that returns tasks as they are ready, but limits the
    number of tasks running at a time.

    Parameters
    ----------
    *coros_or_futures
        The awaitables to run in a bounded concurrent fashion.
    limit : Optional[`int`]
        The maximum number of concurrent tasks. Used when no ``semaphore``
        is passed.
    semaphore : Optional[:class:`asyncio.Semaphore`]
        The semaphore to use for bounding tasks. If `None`, create one
        using ``loop`` and ``limit``.

    Raises
    ------
    TypeError
        When invalid parameters are passed
    """
    loop = asyncio.get_running_loop()

    if semaphore is None:
        if not isinstance(limit, int) or limit <= 0:
            raise TypeError("limit must be an int > 0")

        semaphore = asyncio.Semaphore(limit)

    pending = []

    for cof in coros_or_futures:
        if asyncio.isfuture(cof) and cof._loop is not loop:
            raise ValueError("futures are tied to different event loops")

        cof = _sem_wrapper(semaphore, cof)
        pending.append(cof)

    return asyncio.as_completed(pending)


def bounded_gather(
    *coros_or_futures,
    return_exceptions: bool = False,
    limit: int = 4,
    semaphore: asyncio.Semaphore | None = None,
) -> Awaitable[list[Any]]:
    """
    A semaphore-bounded wrapper to :meth:`asyncio.gather`.

    Parameters
    ----------
    *coros_or_futures
        The awaitables to run in a bounded concurrent fashion.
    return_exceptions : bool
        If true, gather exceptions in the result list instead of raising.
    limit : Optional[`int`]
        The maximum number of concurrent tasks. Used when no ``semaphore``
        is passed.
    semaphore : Optional[:class:`asyncio.Semaphore`]
        The semaphore to use for bounding tasks. If `None`, create one
        using ``loop`` and ``limit``.

    Raises
    ------
    TypeError
        When invalid parameters are passed
    """
    asyncio.get_running_loop()

    if semaphore is None:
        if not isinstance(limit, int) or limit <= 0:
            raise TypeError("limit must be an int > 0")

        semaphore = asyncio.Semaphore(limit)

    tasks = (_sem_wrapper(semaphore, task) for task in coros_or_futures)

    return asyncio.gather(*tasks, return_exceptions=return_exceptions)


class AsyncIter(AsyncIterator[_T], Awaitable[list[_T]]):  # pylint: disable=duplicate-bases
    """Asynchronous iterator yielding items from ``iterable``
    that sleeps for ``delay`` seconds every ``steps`` items.

    Parameters
    ----------
    iterable: Iterable
        The iterable to make async.
    delay: Union[float, int]
        The amount of time in seconds to sleep.
    steps: int
        The number of iterations between sleeps.

    Raises
    ------
    ValueError
        When ``steps`` is lower than 1.

    Examples
    --------
    >>> from pylav.utils import AsyncIter
    >>> async for value in AsyncIter(range(3)):
    ...     print(value)
    0
    1
    2

    """

    def __init__(self, iterable: Iterable[_T], delay: float | int = 0, steps: int = 100) -> None:
        if steps < 1:
            raise ValueError("Steps must be higher than or equals to 1")
        self._delay = delay
        self._iterator = iter(iterable)
        self._i = 0
        self._steps = steps
        self._map = None

    def __aiter__(self) -> AsyncIter[_T]:
        return self

    async def __anext__(self) -> _T:
        try:
            item = next(self._iterator)
        except StopIteration as e:
            raise StopAsyncIteration from e
        if self._i == self._steps:
            self._i = 0
            await asyncio.sleep(self._delay)
        self._i += 1
        return await maybe_coroutine(self._map, item) if self._map is not None else item

    def __await__(self) -> Generator[Any, None, list[_T]]:
        """Returns a list of the iterable.

        Examples
        --------
        >>> from pylav.utils import AsyncIter
        >>> iterator = AsyncIter(range(5))
        >>> await iterator
        [0, 1, 2, 3, 4]

        """
        return self.flatten().__await__()

    async def next(self, default: Any = ...) -> _T:
        """Returns a next entry of the iterable.

        Parameters
        ----------
        default: Optional[Any]
            The value to return if the iterator is exhausted.

        Raises
        ------
        StopAsyncIteration
            When ``default`` is not specified and the iterator has been exhausted.

        Examples
        --------
        >>> from pylav.utils import AsyncIter
        >>> iterator = AsyncIter(range(5))
        >>> await iterator.next()
        0
        >>> await iterator.next()
        1

        """
        try:
            value = await self.__anext__()
        except StopAsyncIteration:
            if default is ...:
                raise
            value = default
        return value

    async def flatten(self) -> list[_T]:
        """Returns a list of the iterable.

        Examples
        --------
        >>> from pylav.utils import AsyncIter
        >>> iterator = AsyncIter(range(5))
        >>> await iterator.flatten()
        [0, 1, 2, 3, 4]

        """
        return [item async for item in self]

    def filter(self, function: Callable[[_T], bool | Awaitable[bool]]) -> AsyncFilter[_T]:
        """Filter the iterable with an (optionally async) predicate.

        Parameters
        ----------
        function: Callable[[T], Union[bool, Awaitable[bool]]]
            A function or coroutine function which takes one item of ``iterable``
            as an argument, and returns ``True`` or ``False``.

        Returns
        -------
        AsyncFilter[T]
            An object which can either be awaited to yield a list of the filtered
            items, or can also act as an async iterator to yield items one by one.

        Examples
        --------
        >>> from pylav.utils import AsyncIter
        >>> def predicate(value):
        ...     return value <= 5
        >>> iterator = AsyncIter([1, 10, 5, 100])
        >>> async for i in iterator.filter(predicate):
        ...     print(i)
        1
        5

        >>> from pylav.utils import AsyncIter
        >>> def predicate(value):
        ...     return value <= 5
        >>> iterator = AsyncIter([1, 10, 5, 100])
        >>> await iterator.filter(predicate)
        [1, 5]

        """
        return async_filter(function, self)

    def enumerate(self, start: int = 0) -> AsyncIterator[tuple[int, _T]]:
        """Async iterable version of `enumerate`.

        Parameters
        ----------
        start: int
            The index to start from. Defaults to 0.

        Returns
        -------
        AsyncIterator[Tuple[int, T]]
            An async iterator of tuples in the form of ``(index, item)``.

        Examples
        --------
        >>> from pylav.utils import AsyncIter
        >>> iterator = AsyncIter(['one', 'two', 'three'])
        >>> async for i in iterator.enumerate(start=10):
        ...     print(i)
        (10, 'one')
        (11, 'two')
        (12, 'three')

        """
        return async_enumerate(self, start)

    async def without_duplicates(self) -> AsyncIterator[_T]:
        """
        Iterates while omitting duplicated entries.

        Examples
        --------
        >>> from pylav.utils import AsyncIter
        >>> iterator = AsyncIter([1,2,3,3,4,4,5])
        >>> async for i in iterator.without_duplicates():
        ...     print(i)
        1
        2
        3
        4
        5

        """
        _temp = set()
        async for item in self:
            if item not in _temp:
                yield item
                _temp.add(item)
        del _temp

    async def find(
        self,
        predicate: Callable[[_T], bool | Awaitable[bool]],
        default: Any | None = None,
    ) -> AsyncIterator[_T]:
        """Calls ``predicate`` over items in iterable and return first value to match.

        Parameters
        ----------
        predicate: Union[Callable, Coroutine]
            A function that returns a boolean-like result. The predicate provided can be a coroutine.
        default: Optional[Any]
            The value to return if there are no matches.

        Raises
        ------
        TypeError
            When ``predicate`` is not a callable.

        Examples
        --------
        >>> from pylav.utils import AsyncIter
        >>> await AsyncIter(range(3)).find(lambda x: x == 1)
        1
        """
        while True:
            try:
                elem = await self.__anext__()
            except StopAsyncIteration:
                return default
            ret = await maybe_coroutine(predicate, elem)
            if ret:
                return elem

    def map(self, func: Callable[[_T], _S | Awaitable[_S]]) -> AsyncIter[_S]:
        """Set the mapping callable for this instance of `AsyncIter`.

        .. important::
            This should be called after AsyncIter initialization and before any other of its methods.

        Parameters
        ----------
        func: Union[Callable, Coroutine]
            The function to map values to. The function provided can be a coroutine.

        Raises
        ------
        TypeError
            When ``func`` is not a callable.

        Examples
        --------
        >>> from pylav.utils import AsyncIter
        >>> async for value in AsyncIter(range(3)).map(bool):
        ...     print(value)
        False
        True
        True

        """

        if not callable(func):
            raise TypeError("Mapping must be a callable.")
        self._map = func
        return self



from __future__ import annotations

from pylav.filters.utils import FilterMixin


class Rotation(FilterMixin):
    def __init__(self, hertz: float):
        self.hertz = hertz
        self.off = False

    def to_dict(self) -> dict:
        return {
            "hertz": self.hertz,
            "off": self.off,
        }

    @classmethod
    def from_dict(cls, data: dict) -> Rotation:
        c = cls(hertz=data["hertz"])
        c.off = data["off"]
        return c

    def __repr__(self):
        return f"<Rotation: hertz={self.hertz}>"

    @property
    def hertz(self) -> float:
        return self._hertz

    @hertz.setter
    def hertz(self, v: float):
        self._hertz = v
        self.off = False

    @classmethod
    def default(cls) -> Rotation:
        c = cls(hertz=0.0)
        c.off = True
        return c

    def get(self) -> dict[str, float]:
        return (
            {}
            if self.off
            else {
                "rotationHz": self.hertz,
            }
        )

    def reset(self) -> None:
        self.hertz = 0.0
        self.off = True



from __future__ import annotations

from pylav.filters.utils import FilterMixin


class Karaoke(FilterMixin):
    def __init__(self, level: float, mono_level: float, filter_band: float, filter_width: float):
        self.level = level
        self.mono_level = mono_level
        self.filter_band = filter_band
        self.filter_width = filter_width
        self.off = False

    def to_dict(self) -> dict:
        return {
            "level": self.level,
            "mono_level": self.mono_level,
            "filter_band": self.filter_band,
            "filter_width": self.filter_width,
            "off": self.off,
        }

    @classmethod
    def from_dict(cls, data: dict) -> Karaoke:
        c = cls(
            level=data["level"],
            mono_level=data["mono_level"],
            filter_band=data["filter_band"],
            filter_width=data["filter_width"],
        )
        c.off = data["off"]
        return c

    def __repr__(self):
        return (
            f"<Karaoke: level={self.level}, mono_level={self.mono_level}, "
            f"filter_band={self.filter_band}, filter_width={self.filter_width}>"
        )

    @property
    def level(self) -> float:
        return self._level

    @level.setter
    def level(self, v: float):
        self._level = v
        self.off = False

    @property
    def mono_level(self) -> float:
        return self._mono_level

    @mono_level.setter
    def mono_level(self, v: float):
        self._mono_level = v
        self.off = False

    @property
    def filter_band(self) -> float:
        return self._filter_band

    @filter_band.setter
    def filter_band(self, v: float):
        self._filter_band = v
        self.off = False

    @property
    def filter_width(self) -> float:
        return self._filter_width

    @filter_width.setter
    def filter_width(self, v: float):
        self._filter_width = v
        self.off = False

    @classmethod
    def default(cls) -> Karaoke:
        c = cls(level=1.0, mono_level=1.0, filter_band=220.0, filter_width=100.0)
        c.off = True
        return c

    def get(self) -> dict[str, float]:
        if self.off:
            return {}
        return {
            "level": self.level,
            "monoLevel": self.mono_level,
            "filterBand": self.filter_band,
            "filterWidth": self.filter_width,
        }

    def reset(self) -> None:
        self.level = 1.0
        self.mono_level = 1.0
        self.filter_band = 220.0
        self.filter_width = 100.0
        self.off = True



from __future__ import annotations

import asyncio
import contextlib
from typing import TYPE_CHECKING, Any

import aiohttp
import ujson

from pylav._logging import getLogger
from pylav.constants import PYLAV_NODES
from pylav.events import (
    SegmentSkippedEvent,
    SegmentsLoadedEvent,
    TrackEndEvent,
    TrackExceptionEvent,
    TrackStartAppleMusicEvent,
    TrackStartBandcampEvent,
    TrackStartClypitEvent,
    TrackStartEvent,
    TrackStartGCTTSEvent,
    TrackStartGetYarnEvent,
    TrackStartHTTPEvent,
    TrackStartLocalFileEvent,
    TrackStartMixCloudEvent,
    TrackStartNicoNicoEvent,
    TrackStartOCRMixEvent,
    TrackStartPornHubEvent,
    TrackStartRedditEvent,
    TrackStartSoundCloudEvent,
    TrackStartSoundgasmEvent,
    TrackStartSpeakEvent,
    TrackStartSpotifyEvent,
    TrackStartTikTokEvent,
    TrackStartTwitchEvent,
    TrackStartVimeoEvent,
    TrackStartYouTubeEvent,
    TrackStartYouTubeMusicEvent,
    TrackStuckEvent,
    WebSocketClosedEvent,
)
from pylav.exceptions import WebsocketNotConnectedError
from pylav.location import get_closest_discord_region
from pylav.node import Stats
from pylav.utils import AsyncIter, ExponentialBackoffWithReset

if TYPE_CHECKING:
    from pylav.client import Client
    from pylav.node import Node
    from pylav.player import Player
    from pylav.tracks import Track

LOGGER = getLogger("PyLav.WebSocket")


def _done_callback(task: asyncio.Task) -> None:
    with contextlib.suppress(asyncio.CancelledError):
        exc = task.exception()
        if exc is not None:
            LOGGER.error("Error in connect task", exc_info=exc)


class WebSocket:
    """Represents the WebSocket connection with Lavalink."""

    def __init__(
        self,
        *,
        node: Node,
        host: str,
        port: int,
        password: str,
        resume_key: str,
        resume_timeout: int,
        reconnect_attempts: int,
        ssl: bool,
    ):
        self._node = node
        self._client = self._node.node_manager.client

        self._session = aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=120), json_serialize=ujson.dumps)
        self._ws = None
        self._message_queue = []
        self._host = host
        self._port = port
        self._password = password
        self._ssl = ssl
        self._max_reconnect_attempts = reconnect_attempts

        self._resume_key = resume_key
        self._resume_timeout = resume_timeout
        self._resuming_configured = False

        self._ws_uri = f"{self.socket_protocol}://{self._host}:{self._port}"

        self._closers = (
            aiohttp.WSMsgType.CLOSE,
            aiohttp.WSMsgType.CLOSING,
            aiohttp.WSMsgType.CLOSED,
        )
        self.ready = asyncio.Event()
        self._connect_task = asyncio.ensure_future(self.connect())
        self._connect_task.add_done_callback(_done_callback)

    @property
    def is_ready(self) -> bool:
        """Returns whether the websocket is ready."""
        return self.ready.is_set() and self.connected

    @property
    def socket_protocol(self) -> str:
        """The protocol used for the socket connection."""
        return "wss" if self._ssl else "ws"

    @property
    def lib_version(self) -> str:
        """Returns the PyLav library version."""
        return self._client.lib_version

    @property
    def bot_id(self) -> str:
        """Returns the bot's ID."""
        return self._client.bot_id

    @property
    def node(self) -> Node:
        """Returns the :class:`Node` instance."""
        return self._node

    @property
    def client(self) -> Client:
        """Returns the :class:`Client` instance."""
        return self._client

    @property
    def connected(self):
        """Returns whether the websocket is connected to Lavalink."""
        return self._ws is not None and not self._ws.closed and self.ready.is_set()

    @property
    def connecting(self):
        """Returns whether the websocket is connecting to Lavalink."""
        return not self.ready.is_set()

    async def ping(self) -> None:
        """Pings the websocket."""
        if self.connected:
            await self._ws.ping()
        else:
            raise WebsocketNotConnectedError

    async def wait_until_ready(self, timeout: float | None = None):
        await asyncio.wait_for(self.ready.wait(), timeout=timeout)

    async def connect(self):  # sourcery no-metrics
        """Attempts to establish a connection to Lavalink."""
        try:
            self.ready.clear()
            self.node._ready.clear()
            headers = {
                "Authorization": self._password,
                "User-Id": str(self.bot_id),
                "Client-Name": f"PyLav/{self.lib_version}",
            }

            if self.client.is_shutting_down:
                return
            if self._resuming_configured and self._resume_key:
                headers["Resume-Key"] = self._resume_key
            if self._node.identifier in PYLAV_NODES:
                # Since these nodes are proxied by Cloudflare - lets add a special case to properly identify them.
                self._node._region, self._node._coordinates = PYLAV_NODES[self._node.identifier]
            else:
                self._node._region, self._node._coordinates = await get_closest_discord_region(self._host)

            is_finite_retry = self._max_reconnect_attempts != -1
            max_attempts_str = self._max_reconnect_attempts if is_finite_retry else "inf"
            attempt = 0
            backoff = ExponentialBackoffWithReset(base=3)
            while not self.connected and (not is_finite_retry or attempt < self._max_reconnect_attempts):
                attempt += 1
                LOGGER.info(
                    "[NODE-%s] Attempting to establish WebSocket connection (%s/%s)...",
                    self._node.name,
                    attempt,
                    max_attempts_str,
                )

                try:
                    self._ws = await self._session.ws_connect(
                        url=self._ws_uri, headers=headers, heartbeat=60, timeout=600
                    )
                    await self._node.update_features()
                    self.ready.set()
                    self.node._ready.set()
                    backoff.reset()
                except (
                    aiohttp.ClientConnectorError,
                    aiohttp.WSServerHandshakeError,
                    aiohttp.ServerDisconnectedError,
                ) as ce:
                    if self.client.is_shutting_down:
                        return
                    if isinstance(ce, aiohttp.ClientConnectorError):
                        LOGGER.warning(
                            "[NODE-%s] Invalid response received; this may indicate that "
                            "Lavalink is not running, or is running on a port different "
                            "to the one you passed to `add_node` (%s - %s)",
                            self.node.name,
                            self._ws_uri,
                            headers,
                        )
                    elif isinstance(ce, aiohttp.WSServerHandshakeError):
                        if ce.status in (
                            401,
                            403,
                        ):  # Special handling for 401/403 (Unauthorized/Forbidden).
                            LOGGER.warning(
                                "[NODE-%s] Authentication failed while trying to establish a connection to the node.",
                                self.node.name,
                            )
                            # We shouldn't try to establish any more connections as correcting this particular error
                            # would require the cog to be reloaded (or the bot to be rebooted), so further attempts
                            # would be futile, and a waste of resources.
                            return

                        LOGGER.warning(
                            "[NODE-%s] The remote server returned code %s, "
                            "the expected code was 101. This usually "
                            "indicates that the remote server is a webserver "
                            "and not Lavalink. Check your ports, and try again.",
                            self.node.name,
                            ce.status,
                        )
                    await asyncio.sleep(backoff.delay())
                else:
                    await self.node.node_manager.node_connect(self.node)
                    #  asyncio.ensure_future(self._listen())
                    if (
                        not self._resuming_configured
                        and self._resume_key
                        and (self._resume_timeout and self._resume_timeout > 0)
                    ):
                        await self.send(
                            op="configureResuming",
                            key=self._resume_key,
                            timeout=self._resume_timeout,
                        )
                        self._resuming_configured = True

                    if self._message_queue:
                        async for message in AsyncIter(self._message_queue):
                            await self.send(**message)

                        self._message_queue.clear()

                    await self._listen()
                    LOGGER.debug("[NODE-%s] _listen returned.", self.node.name)
                    # Ensure this loop doesn't proceed if _listen returns control back to this
                    # function.
                    return

            LOGGER.warning(
                "[NODE-%s] A WebSocket connection could not be established within %s attempts.",
                self.node.name,
                attempt,
            )
        except Exception:
            LOGGER.exception(
                "[NODE-%s] An exception occurred while attempting to connect to the node.",
                self.node.name,
            )

    async def _listen(self):
        """Listens for websocket messages."""
        try:
            async for msg in self._ws:
                LOGGER.trace("[NODE-%s] Received WebSocket message: %s", self.node.name, msg.data)
                if msg.type == aiohttp.WSMsgType.CLOSED:
                    LOGGER.info(
                        "[NODE-%s] Received close frame with code %s.",
                        self.node.name,
                        msg.data,
                    )
                    await self._websocket_closed(msg.data, msg.extra)
                    return
                else:
                    await self.handle_message(msg.json(loads=ujson.loads))
                # elif msg.type == aiohttp.WSMsgType.ERROR and not self.client.is_shutting_down:
                #     exc = self._ws.exception()
                #     LOGGER.error("[NODE-%s] Exception in WebSocket! %s.", self.node.name, exc)
                #     break
            await self._websocket_closed()
        except Exception:
            if not self.client.is_shutting_down:
                LOGGER.exception("[NODE-%s] Exception in WebSocket!", self.node.name)
                await self._websocket_closed()

    async def _websocket_closed(self, code: int = None, reason: str = None):
        """
        Handles when the websocket is closed.

        Parameters
        ----------
        code: :class:`int`
            The response code.
        reason: :class:`str`
            Reason why the websocket was closed. Defaults to `None`
        """
        LOGGER.info(
            "[NODE-%s] WebSocket disconnected with the following: code=%s reason=%s",
            self.node.name,
            code,
            reason,
        )
        self._ws = None
        await self.node.node_manager.node_disconnect(self.node, code, reason)
        if not self._connect_task.cancelled():
            self._connect_task.cancel()
        self._connect_task = asyncio.ensure_future(self.connect())
        self._connect_task.add_done_callback(_done_callback)

    async def handle_message(self, data: dict):
        """
        Handles the response from the websocket.

        Parameters
        ----------
        data: :class:`dict`
            The data given from Lavalink.
        """
        op = data["op"]

        if op == "stats":
            self.node.stats = Stats(self.node, data)
        elif op == "playerUpdate":
            player = self.client.player_manager.get(int(data["guildId"]))

            if not player:
                return

            await player._update_state(data["state"])
        elif op == "event":
            await self.handle_event(data)
        else:
            LOGGER.warning("[NODE-%s] Received unknown op: %s", self.node.name, op)

    async def handle_event(self, data: dict):
        """
        Handles the event from Lavalink.

        Parameters
        ----------
        data: :class:`dict`
            The data given from Lavalink.
        """
        if self.client.is_shutting_down:
            return
        player = self.client.player_manager.get(int(data["guildId"]))
        if not player:
            LOGGER.debug(
                "[NODE-%s] Received event for non-existent player! Guild ID: %s",
                self.node.name,
                data["guildId"],
            )
            return

        event_type = data["type"]
        if event_type == "TrackEndEvent":
            from pylav.query import Query
            from pylav.tracks import Track

            requester = None
            track = None
            if player.current and player.current.track == data["track"]:
                player.current.timestamp = 0
                requester = player.current.requester
                track = player.current

            event = TrackEndEvent(
                player,
                Track(
                    data=data["track"],
                    requester=requester.id if requester else self._client.bot.user.id,
                    query=await Query.from_base64(data["track"]),
                    node=self.node,
                )
                if not track
                else track,
                data["reason"],
                self.node,
            )
            await player._handle_event(event)
        elif event_type == "TrackExceptionEvent":
            event = TrackExceptionEvent(player, player.current, data["error"], node=self.node)
            if self.node.identifier == player.node.identifier:
                await player._handle_event(event)
        elif event_type == "TrackStartEvent":
            track = player.current
            event = TrackStartEvent(player, track, self.node)
            await self._process_track_event(player, track, self.node)
        elif event_type == "TrackStuckEvent":
            event = TrackStuckEvent(player, player.current, data["thresholdMs"], self.node)
            await player._handle_event(event)
        elif event_type == "WebSocketClosedEvent":
            event = WebSocketClosedEvent(
                player, data["code"], data["reason"], data["byRemote"], self.node, player.channel
            )
        elif event_type == "SegmentsLoaded":
            event = SegmentsLoadedEvent(player, data["segments"], self.node)
        elif event_type == "SegmentSkipped":
            event = SegmentSkippedEvent(player, node=self.node, **data["segment"])
        else:
            LOGGER.warning("[NODE-%s] Unknown event received: %s", self.node.name, event_type)
            return

        self.client.dispatch_event(event)

    async def send(self, **data: Any):
        """
        Sends a payload to Lavalink.

        Parameters
        ----------
        data: :class:`dict`
            The data sent to Lavalink.
        """
        if self.connected:
            LOGGER.trace("[NODE-%s] Sending payload %s", self.node.name, data)
            try:
                await self._ws.send_json(data)
            except ConnectionResetError:
                LOGGER.debug("[NODE-%s] Send called before WebSocket ready!", self.node.name)
                self._message_queue.append(data)
        else:
            LOGGER.debug("[NODE-%s] Send called before WebSocket ready!", self.node.name)
            self._message_queue.append(data)

    async def _process_track_event(self, player: Player, track: Track, node: Node) -> None:
        # sourcery no-metrics
        if await track.is_youtube_music():
            event = TrackStartYouTubeMusicEvent(player, track, node)
        elif await track.is_spotify():
            event = TrackStartSpotifyEvent(player, track, node)
        elif await track.is_apple_music():
            event = TrackStartAppleMusicEvent(player, track, node)
        elif await track.is_local():
            event = TrackStartLocalFileEvent(player, track, node)
        elif await track.is_http():
            event = TrackStartHTTPEvent(player, track, node)
        elif await track.is_speak():
            event = TrackStartSpeakEvent(player, track, node)
        elif await track.is_youtube():
            event = TrackStartYouTubeEvent(player, track, node)
        elif await track.is_clypit():
            event = TrackStartClypitEvent(player, track, node)
        elif await track.is_getyarn():
            event = TrackStartGetYarnEvent(player, track, node)
        elif await track.is_twitch():
            event = TrackStartTwitchEvent(player, track, node)
        elif await track.is_vimeo():
            event = TrackStartVimeoEvent(player, track, node)
        elif await track.is_mixcloud():
            event = TrackStartMixCloudEvent(player, track, node)
        elif await track.is_ocremix():
            event = TrackStartOCRMixEvent(player, track, node)
        elif await track.is_pornhub():
            event = TrackStartPornHubEvent(player, track, node)
        elif await track.is_reddit():
            event = TrackStartRedditEvent(player, track, node)
        elif await track.is_soundgasm():
            event = TrackStartSoundgasmEvent(player, track, node)
        elif await track.is_tiktok():
            event = TrackStartTikTokEvent(player, track, node)
        elif await track.is_bandcamp():
            event = TrackStartBandcampEvent(player, track, node)
        elif await track.is_soundcloud():
            event = TrackStartSoundCloudEvent(player, track, node)
        elif await track.is_gctts():
            event = TrackStartGCTTSEvent(player, track, node)
        elif await track.is_niconico():
            event = TrackStartNicoNicoEvent(player, track, node)
        else:  # This should never happen
            event = TrackStartEvent(player, track, node)
        self.client.dispatch_event(event)

    async def close(self):
        self._connect_task.cancel()
        if self._ws and not self._ws.closed:
            await self._ws.close(code=4014, message=b"Shutting down...")
        await self._session.close()



from __future__ import annotations

import secrets

from pylav._config import __VERSION__

NODE_DEFAULT_SETTINGS = {
    "server": {"port": 2154, "address": "localhost"},
    "lavalink": {
        "plugins": [
            {
                "dependency": "com.github.Topis-Lavalink-Plugins:Topis-Source-Managers-Plugin:v2.0.7",
                "repository": "https://jitpack.io",
            },
            {
                "dependency": "com.dunctebot:skybot-lavalink-plugin:1.4.0",
                "repository": "https://m2.duncte123.dev/releases",
            },
            {"dependency": "com.github.topisenpai:sponsorblock-plugin:v1.0.3", "repository": "https://jitpack.io"},
        ],
        "server": {
            "password": secrets.token_urlsafe(32),
            "sources": {
                "youtube": True,
                "bandcamp": True,
                "soundcloud": True,
                "twitch": True,
                "vimeo": True,
                "http": True,
                "local": True,
            },
            "bufferDurationMs": 400,
            "frameBufferDurationMs": 1000,
            "trackStuckThresholdMs": 10000,
            "youtubePlaylistLoadLimit": 100,
            "opusEncodingQuality": 10,
            "resamplingQuality": "LOW",
            "useSeekGhosting": True,
            "playerUpdateInterval": 1,
            "youtubeSearchEnabled": True,
            "soundcloudSearchEnabled": True,
            "gc-warnings": True,
            "ratelimit": {
                "ipBlocks": [],
                "excludedIps": [],
                "strategy": "RotateOnBan",
                "searchTriggersFail": True,
                "retryLimit": -1,
            },
            "youtubeConfig": {
                "email": "",
                "password": "",
            },
            "httpConfig": {"proxyHost": "", "proxyPort": 0, "proxyUser": "", "proxyPassword": ""},
        },
    },
    "plugins": {
        "topissourcemanagers": {
            "providers": [
                'ytmsearch:"%ISRC%"',
                'ytsearch:"%ISRC%"',
                "ytmsearch:%QUERY%",
                "ytsearch:%QUERY%",
                "scsearch:%QUERY%",
            ],
            "sources": {"spotify": True, "applemusic": True},
            "spotify": {
                "clientId": "3d5cd36c73924786aa290798b2131c58",
                "clientSecret": "edee5eb255a846fbac8297069debea2e",
                "countryCode": "US",
            },
            "applemusic": {"countryCode": "US"},
        },
        "dunctebot": {
            "ttsLanguage": "en-US",
            "sources": {
                "getyarn": True,
                "clypit": True,
                "tts": True,
                "pornhub": True,
                "reddit": True,
                "ocremix": True,
                "tiktok": True,
                "mixcloud": True,
                "soundgasm": True,
            },
        },
    },
    "metrics": {"prometheus": {"enabled": False, "endpoint": "/metrics"}},
    "sentry": {"dsn": "", "environment": "", "tags": {"pylav_version": __VERSION__}},
    "logging": {
        "file": {"path": "./logs/"},
        "logback": {
            "rollingpolicy": {"max-history": 7, "max-file-size": "25MB", "total-size-cap": "1GB"},
        },
        "level": {"root": "INFO", "lavalink": "INFO"},
    },
}



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, TypeVar

from discord.app_commands import Choice, Transformer
from discord.ext import commands

from pylav.types import ContextT, InteractionT

try:
    from redbot.core.i18n import Translator

    _ = Translator("PyLavPlayer", Path(__file__))
except ImportError:
    _ = lambda x: x

if TYPE_CHECKING:
    from pylav.sql.models import PlaylistModel

    PlaylistConverter = TypeVar("PlaylistConverter", bound=list[PlaylistModel])
else:

    class PlaylistConverter(Transformer):
        @classmethod
        async def convert(cls, ctx: ContextT, arg: str) -> list[PlaylistModel]:
            """Converts a playlist name or ID to a list of matching objects."""
            from pylav import EntryNotFoundError

            try:
                playlists = await ctx.lavalink.playlist_db_manager.get_playlist_by_name_or_id(arg)
            except EntryNotFoundError as e:
                raise commands.BadArgument(_("Playlist with name or id `{arg}` not found.").format(arg=arg)) from e
            return playlists

        @classmethod
        async def transform(cls, interaction: InteractionT, argument: str) -> list[PlaylistModel]:
            if not interaction.response.is_done():
                await interaction.response.defer(ephemeral=True)
            ctx = await interaction.client.get_context(interaction)
            return await cls.convert(ctx, argument)

        @classmethod
        async def autocomplete(cls, interaction: InteractionT, current: str) -> list[Choice]:
            return [
                Choice(name=p.name[:99], value=f"{p.id}")
                for p in await interaction.client.lavalink.playlist_db_manager.get_playlist_by_name(current, limit=25)
                if current.lower() in p.name.lower()
            ]



from __future__ import annotations

import asyncio
import operator
import os
from functools import partial
from typing import TYPE_CHECKING

import aiohttp
import asyncstdlib
import ujson

from pylav._logging import getLogger
from pylav.constants import (
    BUNDLED_NODES_IDS,
    DEFAULT_REGIONS,
    PYLAV_BUNDLED_NODES_SETTINGS,
    REGION_TO_COUNTRY_COORDINATE_MAPPING,
)
from pylav.envvars import JAVA_EXECUTABLE, USE_BUNDLED_EXTERNAL_LAVA_LINK_NODE, USE_BUNDLED_EXTERNAL_PYLAV_NODE
from pylav.events import NodeConnectedEvent, NodeDisconnectedEvent
from pylav.location import get_closest_region_name_and_coordinate
from pylav.node import Node
from pylav.player import Player
from pylav.sql.models import NodeModel
from pylav.utils import sort_key_nodes

if TYPE_CHECKING:
    from pylav.client import Client

LOGGER = getLogger("PyLav.NodeManager")


class NodeManager:
    def __init__(
        self,
        client: Client,
        external_host: str = None,
        external_password: str = None,
        external_port: int = None,
        external_ssl: bool = False,
    ):
        self._client = client
        self._session = aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=30), json_serialize=ujson.dumps)
        self._player_queue = []
        self._unmanaged_external_host = external_host
        self._unmanaged_external_password = external_password
        self._unmanaged_external_port = external_port
        self._unmanaged_external_ssl = external_ssl
        self._nodes = []
        self._adding_nodes = asyncio.Event()

    def __iter__(self):
        yield from self._nodes

    @property
    def session(self) -> aiohttp.ClientSession:
        return self._session

    @property
    def client(self) -> Client:
        """Returns the client."""
        return self._client

    @property
    def nodes(self) -> list[Node]:
        """Returns a list of all nodes."""
        return self._nodes

    @property
    def available_nodes(self) -> list[Node]:
        """Returns a list of available nodes."""
        return list(filter(operator.attrgetter("available"), self.nodes))

    @property
    def managed_nodes(self) -> list[Node]:
        """Returns a list of nodes that are managed by the client."""
        return list(filter(operator.attrgetter("managed"), self.nodes))

    @property
    def search_only_nodes(self) -> list[Node]:
        """Returns a list of nodes that are search only."""
        return list(filter(operator.attrgetter("available", "search_only"), self.nodes))

    @property
    def player_queue(self) -> list[Player]:
        """Returns a list of players that are queued to be played."""
        return self._player_queue

    @player_queue.setter
    def player_queue(self, players: list[Player]) -> None:
        """Sets the player queue."""
        self._player_queue = players

    @player_queue.deleter
    def player_queue(self):
        """Clears the player queue."""
        self._player_queue.clear()

    async def add_node(
        self,
        *,
        host: str,
        port: int,
        password: str,
        unique_identifier: int,
        name: str,
        resume_key: str = None,
        resume_timeout: int = 60,
        reconnect_attempts: int = -1,
        ssl: bool = False,
        search_only: bool = False,
        skip_db: bool = False,
        disabled_sources: list[str] = None,
        managed: bool = False,
        yaml: dict | None = None,
        extras: dict = None,
    ) -> Node:
        """
        Adds a node to PyLav's node manager.

        Parameters
        ----------
        host: :class:`str`
            The address of the Lavalink node.
        port: :class:`int`
            The port to use for websocket and REST connections.
        password: :class:`str`
            The password used for authentication.
        resume_key: Optional[:class:`str`]
            A resume key used for resuming a session upon re-establishing a WebSocket connection to Lavalink.
            Defaults to `None`.
        resume_timeout: Optional[:class:`int`]
            How long the node should wait for a connection while disconnected before clearing all players.
            Defaults to `60`.
        name: :class:`str`
            An identifier for the node that will show in logs. Defaults to `None`.
        reconnect_attempts: Optional[:class:`int`]
            The amount of times connection with the node will be reattempted before giving up.
            Set to `-1` for infinite. Defaults to `3`.
        ssl: Optional[:class:`bool`]
            Whether to use a ssl connection. Defaults to `False`.
        search_only: :class:`bool`
            Whether the node is search only. Defaults to `False`.
        unique_identifier: Optional[:class:`str`]
            A unique identifier for the node. Defaults to `None`.
        skip_db: Optional[:class:`bool`]
            Whether to skip the database creation op. Defaults to `False`.
        disabled_sources: Optional[:class:`list`[:class:`str`]]
            A list of sources to disable. Defaults to `None`.
        managed: Optional[:class:`bool`]
            Whether the node is managed by the client. Defaults to `False`.
        yaml: Optional[:class:`dict`]
            A dictionary of node settings. Defaults to `None`.
        extras: Optional[:class:`dict`]
            A dictionary of extra settings. Defaults to `{}`.

        Returns
        -------
        :class:`Node`
            The node that was added.
        """
        node = Node(
            manager=self,
            host=host,
            port=port,
            password=password,
            resume_key=resume_key,
            resume_timeout=resume_timeout,
            name=name,
            reconnect_attempts=reconnect_attempts,
            ssl=ssl,
            search_only=search_only,
            unique_identifier=unique_identifier,
            disabled_sources=disabled_sources,
            managed=managed,
            extras=extras or {},
        )
        self._nodes.append(node)

        LOGGER.info("[NODE-%s] Successfully added to Node Manager", node.name)
        LOGGER.verbose("[NODE-%s] Successfully added to Node Manager -- %r", node.name, node)
        if skip_db:
            data = dict(
                yaml=yaml or {"server": {}, "lavalink": {"server": {}}},
                id=unique_identifier,
                ssl=ssl,
                reconnect_attempts=reconnect_attempts,
                search_only=search_only,
                resume_key=resume_key,
                resume_timeout=resume_timeout,
                managed=managed,
                name=name,
                disabled_sources=disabled_sources or [],
                extras=extras or {},
            )
            data["yaml"]["server"]["address"] = host  # type: ignore
            data["yaml"]["server"]["port"] = port  # type: ignore
            data["yaml"]["lavalink"]["server"]["password"] = password

            node_config = NodeModel(**data)

        node._config = (
            node_config
            if skip_db
            else await self.client.node_db_manager.add_node(
                host=host,
                port=port,
                password=password,
                resume_key=resume_key,
                resume_timeout=resume_timeout,
                name=name,
                reconnect_attempts=reconnect_attempts,
                ssl=ssl,
                search_only=search_only,
                unique_identifier=unique_identifier,
                disabled_sources=disabled_sources,
                managed=managed,
                yaml=yaml,
                extras=extras or {},
            )
        )

        return node

    async def remove_node(self, node: Node) -> None:
        """
        Removes a node.
        Parameters
        ----------
        node: :class:`Node`
            The node to remove from the list.
        """
        await node.close()
        self.nodes.remove(node)
        LOGGER.info("[NODE-%s] Successfully removed Node", node.name)
        LOGGER.verbose("[NODE-%s] Successfully removed Node -- %r", node.name, node)
        if node.identifier and not node.managed:
            await self.client.node_db_manager.delete(node.identifier)
            LOGGER.debug("[NODE-%s] Successfully deleted Node from database", node.name)

    def get_region(self, endpoint: str | None) -> str | None:
        """
        Returns a region from a Discord voice server address.
        Parameters
        ----------
        endpoint: :class:`str`
            The address of the Discord voice server.
        Returns
        -------
        Optional[:class:`str`]
        """
        if not endpoint:
            return None

        endpoint = endpoint.replace("vip-", "")

        for key in DEFAULT_REGIONS:
            nodes = [n for n in self.available_nodes if n.region == key]

            if not nodes:
                continue

            if endpoint.startswith(key):
                return key
        return None

    def get_closest_node(self, region: str) -> Node:
        """
        Returns the closest node to a given region.
        Parameters
        ----------
        region: :class:`str`
            The region to use.
        Returns
        -------
        :class:`Node`
        """
        return min(self.available_nodes, key=lambda n: n.region_distance(region))

    async def find_best_node(
        self,
        region: str = None,
        not_region: str = None,
        feature: str = None,
        already_attempted_regions: set[str] = None,
        coordinates: tuple[float, float] = None,
    ) -> Node | None:
        """Finds the best (least used) node in the given region, if applicable.
        Parameters
        ----------
        region: :class:`str`
            The region to use.
        not_region: :class:`str`
            The region to exclude.
        feature: :class:`str`
            The feature required.
        already_attempted_regions: :class:`set`[:class:`str`]
            A set of regions that have already been attempted.
        coordinates: :class:`tuple`[:class:`float`, :class:`float`]
            The coordinates to use.
        Returns
        -------
        Optional[:class:`Node`]
        """
        already_attempted_regions = already_attempted_regions or set()
        if feature:
            nodes = [n for n in self.available_nodes if n.has_capability(feature)]
        else:
            nodes = self.available_nodes
        if coordinates is None:
            if region and region in REGION_TO_COUNTRY_COORDINATE_MAPPING:
                coordinates = REGION_TO_COUNTRY_COORDINATE_MAPPING[region]
            else:
                coordinates = (0, 0)
        if region and not_region:
            available_regions = {n.region for n in self.available_nodes if n.region not in already_attempted_regions}
            closest_region, __ = await get_closest_region_name_and_coordinate(
                *coordinates, region_pool=available_regions
            )
            nodes = [
                n
                for n in nodes
                if (n.region in [region, closest_region])
                and n.region != not_region
                and n.region not in already_attempted_regions
            ]
        elif region:
            available_regions = {n.region for n in self.available_nodes if n.region not in already_attempted_regions}
            closest_region, __ = await get_closest_region_name_and_coordinate(
                *coordinates, region_pool=available_regions
            )
            nodes = [
                n for n in nodes if (n.region in [region, closest_region]) and n.region not in already_attempted_regions
            ]
        else:
            nodes = [n for n in nodes if n.region != not_region and n.region not in already_attempted_regions]

        if not nodes:
            if feature:
                nodes = [
                    n
                    for n in self.available_nodes
                    if n.has_capability(feature) and n.region not in already_attempted_regions
                ]
            else:
                nodes = self.available_nodes
        return await asyncstdlib.min(nodes, key=partial(sort_key_nodes, region=region), default=None) if nodes else None

    def get_node_by_id(self, unique_identifier: int) -> Node | None:
        """
        Returns a node by its unique identifier.
        Parameters
        ----------
        unique_identifier: :class:`int`
            The unique identifier of the node.
        Returns
        -------
        Optional[:class:`Node`]
        """
        return next((n for n in self.nodes if n.identifier == unique_identifier), None)

    async def node_connect(self, node: Node) -> None:
        """
        Called when a node is connected from Lavalink.
        Parameters
        ----------
        node: :class:`Node`
            The node that has just connected.
        """
        LOGGER.info("[NODE-%s] Successfully established connection", node.name)

        for player in self.player_queue:
            await player.change_node(node)
            LOGGER.debug("[NODE-%s] Successfully moved %s", node.name, player.guild_id)

        if self.client._connect_back:
            for player in node._original_players:
                await player.change_node(node)
                player._original_node = None

        del self.player_queue
        self.client.dispatch_event(NodeConnectedEvent(node))

    async def node_disconnect(self, node: Node, code: int, reason: str) -> None:
        """
        Called when a node is disconnected from Lavalink.
        Parameters
        ----------
        node: :class:`Node`
            The node that has just connected.
        code: :class:`int`
            The code for why the node was disconnected.
        reason: :class:`str`
            The reason why the node was disconnected.
        """
        if self.client.is_shutting_down:
            return
        LOGGER.warning("[NODE-%s] Disconnected with code %s and reason %s", node.name, code, reason)
        LOGGER.verbose(
            "[NODE-%s] Disconnected with code %s and reason %s -- %r",
            node.name,
            code,
            reason,
            node,
        )
        self.client.dispatch_event(NodeDisconnectedEvent(node, code, reason))
        best_node = await self.find_best_node(region=node.region)

        if not best_node:
            self.player_queue.extend(node.players)
            LOGGER.error("Unable to move players, no available nodes! Waiting for a node to become available.")
            return

        for player in node.players:
            await player.change_node(best_node)

            if self.client._connect_back:
                player._original_node = node

    async def close(self) -> None:
        await self.session.close()
        for node in self.nodes:
            await node.close()

    async def connect_to_all_nodes(self) -> None:
        nodes_list = []
        for node in await self.client.node_db_manager.get_all_unamanaged_nodes():
            try:
                if node in nodes_list:
                    LOGGER.warning(
                        "%s Node already added to connection pool - skipping duplicated connection - (%s:%s)",
                        node.name,
                        node.yaml["server"]["address"],
                        node.yaml["server"]["port"],
                    )
                    continue
                if node.yaml["server"]["address"] in PYLAV_BUNDLED_NODES_SETTINGS:
                    connection_arguments = PYLAV_BUNDLED_NODES_SETTINGS[node.yaml["server"]["address"]]
                else:
                    connection_arguments = node.get_connection_args()
                    connection_arguments["skip_db"] = node.id not in BUNDLED_NODES_IDS
                nodes_list.append(await self.add_node(**connection_arguments))
            except (ValueError, KeyError) as exc:
                LOGGER.warning(
                    "[NODE-%s] Invalid node, skipping ... id: %s - Original error: %s", node.name, node.id, exc
                )
                continue
        if self._unmanaged_external_host and self._unmanaged_external_password:
            if await asyncstdlib.all(True for n in nodes_list if n.host != self._unmanaged_external_host):
                if self._unmanaged_external_host in PYLAV_BUNDLED_NODES_SETTINGS:
                    base_settings = PYLAV_BUNDLED_NODES_SETTINGS[self._unmanaged_external_host]
                else:
                    base_settings = {
                        "port": self._unmanaged_external_port or (443 if self._unmanaged_external_ssl else 80),
                        "ssl": self._unmanaged_external_ssl,
                        "password": self._unmanaged_external_password,
                        "resume_timeout": 600,
                        "reconnect_attempts": -1,
                        "search_only": False,
                        "managed": False,
                        "disabled_sources": [],
                        "skip_db": True,
                        "host": self._unmanaged_external_host,
                        "unique_identifier": 31415,
                        "name": "ENVAR Node (Unmanaged)",
                        "resume_key": f"PyLav/{self.client.lib_version}-{self.client.bot_id}",
                    }
                nodes_list.append(await self.add_node(**base_settings))
            else:
                LOGGER.warning(
                    "ENVAR Node (Unmanaged) already added to connection pool - skipping duplicated connection - (%s:%s)",
                    self._unmanaged_external_host,
                    self._unmanaged_external_port,
                )
        config_data = await self.client._lib_config_manager.get_config(
            config_folder=self.client._config_folder,
            java_path=JAVA_EXECUTABLE,
            enable_managed_node=True,
            auto_update_managed_nodes=True,
            localtrack_folder=self.client._config_folder / "music",
            use_bundled_pylav_external=USE_BUNDLED_EXTERNAL_PYLAV_NODE,
            use_bundled_lava_link_external=USE_BUNDLED_EXTERNAL_LAVA_LINK_NODE,
        )
        if config_data.java_path != JAVA_EXECUTABLE and JAVA_EXECUTABLE != "java" and os.path.exists(JAVA_EXECUTABLE):
            config_data.java_path = JAVA_EXECUTABLE
            await config_data.save()
        if config_data.use_bundled_pylav_external:
            if await asyncstdlib.all(True for n in nodes_list if n.host != "ll-gb.draper.wtf"):
                base_settings = PYLAV_BUNDLED_NODES_SETTINGS["ll-gb.draper.wtf"]
                base_settings["host"] = "ll-gb.draper.wtf"
                base_settings["resume_key"] = f"PyLav/{self.client.lib_version}-{self.client.bot_id}"
                nodes_list.append(await self.add_node(**base_settings))
            else:
                LOGGER.debug(
                    "%s already added to connection pool - skipping duplicated connection",
                    PYLAV_BUNDLED_NODES_SETTINGS["ll-gb.draper.wtf"]["name"],
                )

            if await asyncstdlib.all(True for n in nodes_list if n.host != "ll-us-ny.draper.wtf"):
                base_settings = PYLAV_BUNDLED_NODES_SETTINGS["ll-us-ny.draper.wtf"]
                base_settings["resume_key"] = f"PyLav/{self.client.lib_version}-{self.client.bot_id}"
                nodes_list.append(await self.add_node(**base_settings))
            else:
                LOGGER.debug(
                    "%s already added to connection pool - skipping duplicated connection",
                    PYLAV_BUNDLED_NODES_SETTINGS["ll-us-ny.draper.wtf"]["name"],
                )
        if config_data.use_bundled_lava_link_external:
            if await asyncstdlib.all(True for n in nodes_list if n.host != "lava.link"):
                nodes_list.append(
                    await self.add_node(
                        password=f"PyLav/{self.client.lib_version}",
                        resume_key=f"PyLav/{self.client.lib_version}-{self.client.bot_id}",
                        **PYLAV_BUNDLED_NODES_SETTINGS["lava.link"],
                    )
                )
            else:
                LOGGER.debug(
                    "%s already added to connection pool - skipping duplicated connection",
                    PYLAV_BUNDLED_NODES_SETTINGS["lava.link"]["name"],
                )
        tasks = [asyncio.create_task(n.wait_until_ready()) for n in nodes_list]
        if (
            not (
                self.client._node_manager._unmanaged_external_password
                and self.client._node_manager._unmanaged_external_host
            )
        ) and self.client.enable_managed_node:
            tasks.append(asyncio.create_task(self.client._local_node_manager.wait_until_connected()))
        if not tasks:
            LOGGER.warning("No nodes found, please add some nodes.")
            return
        done, pending = await asyncio.wait(tasks, return_when=asyncio.FIRST_COMPLETED)
        for task in pending:
            task.cancel()
        for result in done:
            result.result()
        if not self._adding_nodes.is_set():
            self._adding_nodes.set()

    async def wait_until_ready(self, timeout: float | None = None):
        await asyncio.wait_for(self._adding_nodes.wait(), timeout=timeout)
